from __future__ import annotations

import numpy as np
from ase.units import Ha
from scipy.special import erfcx

K_G = 8 * np.sqrt(2) / (3 * np.pi**2)  # 0.382106112167171


class Coefficients:
    """Coefficient calculator for GLLB functionals.

    This class implements the calculation of sqrt(E) coefficients as given by
    Eq. (16) of https://doi.org/10.1103/PhysRevB.82.115106.

    Parameters
    ----------
    eps (in eV)
        This parameter cuts sqrt(E) to zero for E < eps.
        The cut is supposed to help convergence with degenerate systems.
        This parameter should be small.
    width (in eV)
        If this parameter is set, then a smoothed variant of the sqrt(E)
        expression is used. This parameter sets the energy width of
        the smoothing.
        The eps parameter is ignored when the width parameter is set.
    """
    def __init__(self,
                 eps: float = 0.05,
                 width: float = None):
        self.width: float | None
        if width is not None:
            self.width = width / Ha
            self.eps = None  # Make sure that eps is not used with width
        else:
            self.width = None
            self.eps = eps / Ha

    def initialize(self, wfs):
        self.wfs = wfs

    def initialize_1d(self, ae):
        self.ae = ae

    def get_description(self):
        desc = []
        if self.eps is not None:
            desc += ['eps={:.4f} eV'.format(self.eps * Ha)]
        if self.width is not None:
            desc += ['width={:.4f} eV'.format(self.width * Ha)]
        return ', '.join(desc)

    def f(self, energy_n):
        """Calculate the sqrt(E)-like coefficient.

        See the class description for details.
        """
        w_n = np.zeros_like(energy_n, dtype=float)
        if self.width is None:
            flt_n = energy_n > self.eps
            w_n[flt_n] = np.sqrt(energy_n[flt_n])
        else:
            prefactor = 0.5 * np.sqrt(np.pi * self.width)
            rel_energy_n = energy_n / self.width
            # Evaluate positive energies
            flt_n = energy_n > 0.0
            w_n[flt_n] = (np.sqrt(energy_n[flt_n])
                          + prefactor * erfcx(np.sqrt(rel_energy_n[flt_n])))
            # Evaluate negative energies
            flt_n = np.logical_not(flt_n)
            w_n[flt_n] = prefactor * np.exp(rel_energy_n[flt_n])
        return w_n

    def get_reference_homo_1d(self):
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        homo = np.max(e_j[f_j > 1e-3])
        return homo

    def get_reference_lumo_1d(self):
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        lumo = np.min(e_j[f_j < 1e-3])
        return lumo

    def get_coefficients_1d(self, smooth=False):
        homo = self.get_reference_homo_1d()
        if smooth:
            w_ln = []
            for e_n, f_n in zip(self.ae.e_ln, self.ae.f_ln):
                e_n = np.asarray(e_n)
                f_n = np.asarray(f_n)
                w_n = f_n * K_G * self.f(homo - e_n)
                w_ln.append(w_n)
            return w_ln
        else:
            e_j = np.asarray(self.ae.e_j)
            f_j = np.asarray(self.ae.f_j)
            return f_j * K_G * self.f(homo - e_j)

    def get_coefficients_1d_for_lumo_perturbation(self, smooth=False):
        homo = self.get_reference_homo_1d()
        lumo = self.get_reference_lumo_1d()
        e_j = np.asarray(self.ae.e_j)
        f_j = np.asarray(self.ae.f_j)
        return f_j * K_G * (self.f(lumo - e_j) - self.f(homo - e_j))

    def get_coefficients(self, kpt_u, eref_s):
        w_kn = []
        for kpt in kpt_u:
            w_n = self.f(eref_s[kpt.s] - kpt.eps_n)
            w_n *= kpt.f_n * K_G
            w_kn.append(w_n)
        return w_kn

    def get_coefficients_for_lumo_perturbation(self, kpt_u, eref_s,
                                               eref_lumo_s):
        w_kn = []
        for kpt in kpt_u:
            w_n = (self.f(eref_lumo_s[kpt.s] - kpt.eps_n)
                   - self.f(eref_s[kpt.s] - kpt.eps_n))
            w_n *= kpt.f_n * K_G
            w_kn.append(w_n)
        return w_kn



from gpaw.lcaotddft.observer import TDDFTObserver


class RestartFileWriter(TDDFTObserver):
    """Observer for writing restart files periodically.

    At the given interval, the calculator restart file is written and
    ``write_restart()`` of every attached observer is called.

    The observer attaches to the TDDFT calculator during creation.

    Parameters
    ----------
    paw
        TDDFT calculator
    restart_filename
        File for writing the calculator object
    interval
        Update interval. The restart files are written every
        that many propagation steps.
    """
    def __init__(self, paw, restart_filename, interval=100):
        TDDFTObserver.__init__(self, paw, interval)
        self.restart_filename = restart_filename

    def _update(self, paw):
        if paw.niter == 0:
            return
        paw.log('%s activated' % self.__class__.__name__)
        for obs, n, args, kwargs in paw.observers:
            if isinstance(obs, TDDFTObserver):
                obs.write_restart()
        paw.write(self.restart_filename, mode='all')



from gpaw.transformers import Transformer
import numpy.random as ra
from gpaw.grid_descriptor import GridDescriptor


def test_fd_ops_non_periodic():
    p = 0
    n = 20
    gd1 = GridDescriptor((n, n, n), (8.0, 8.0, 8.0), pbc_c=p)
    a1 = gd1.zeros()
    ra.seed(8)
    a1[:] = ra.random(a1.shape)
    gd2 = gd1.refine()
    a2 = gd2.zeros()
    i = Transformer(gd1, gd2).apply
    i(a1, a2)
    assert abs(gd1.integrate(a1) - gd2.integrate(a2)) < 1e-10
    r = Transformer(gd2, gd1).apply
    a2[0] = 0.0
    a2[:, 0] = 0.0
    a2[:, :, 0] = 0.0
    a2[-1] = 0.0
    a2[:, -1] = 0.0
    a2[:, :, -1] = 0.0
    r(a2, a1)
    assert abs(gd1.integrate(a1) - gd2.integrate(a2)) < 1e-10



import pytest
import numpy as np
from gpaw.spline import Spline
from gpaw.poisson import FDPoissonSolver, FFTPoissonSolver, PoissonSolver
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC


@pytest.mark.ci
def test_poisson_poisson():
    L = 2.87 / 0.529177

    def f(n, p):
        N = 2 * n
        gd = GridDescriptor((N, N, N), (L, L, L))
        a = gd.zeros()
        print(a.shape)
        # p = PoissonSolver(nn=1, relax=relax)
        p.set_grid_descriptor(gd)
        cut = N / 2.0 * 0.9
        s = Spline(l=0, rmax=cut, f_g=np.array([1, 0.5, 0.0]))
        c = LFC(gd, [[s], [s]])
        c.set_positions([(0, 0, 0), (0.5, 0.5, 0.5)])
        c.add(a)

        I0 = gd.integrate(a)
        a -= I0 / L**3

        b = gd.zeros()
        p.solve(b, a, charge=0)
        return gd.collect(b, broadcast=1)

    b1 = f(8, FDPoissonSolver(nn=1, relax='J', eps=1e-20))
    b2 = f(8, FDPoissonSolver(nn=1, relax='GS', eps=1e-20))
    b3 = f(8, PoissonSolver(nn=1))
    err1 = abs(b1[0, 0, 0] - b1[8, 8, 8])
    err2 = abs(b2[0, 0, 0] - b2[8, 8, 8])
    err3 = abs(b1[0, 0, 0] - b3[8, 8, 8])
    print(err1)
    print(err2)
    print(err3)
    assert err3 < 1e-10
    assert err1 < 8e-16
    assert err2 < 3e-6  # XXX Shouldn't this be better?

    b4 = f(8, FFTPoissonSolver())
    err4 = abs(b4[0, 0, 0] - b4[8, 8, 8])
    print(err4)
    assert err4 < 6e-16



"""Check that our tab-completion script has been updated."""
from ase.cli.completion import update
from gpaw.cli.completion import path
from gpaw.cli.main import commands


def test_complete():
    try:
        update(path, commands, test=True)
    except ValueError:
        raise ValueError(
            'Please update gpaw/cli/complete.py using '
            '"python3 -m gpaw.test.test_complete".')


if __name__ == '__main__':
    update(path, commands, test=False)



from functools import partial

import numpy as np
from gpaw.utilities.grid_redistribute import general_redistribute
from gpaw.utilities.partition import AtomPartition, AtomicMatrixDistributor


class GridRedistributor:
    def __init__(self, comm, broadcast_comm, gd, aux_gd):
        self.comm = comm
        self.broadcast_comm = broadcast_comm
        self.gd = gd
        self.aux_gd = aux_gd
        self.enabled = np.any(gd.parsize_c != aux_gd.parsize_c)

        assert gd.comm.size * broadcast_comm.size == comm.size
        if self.enabled:
            assert comm.compare(aux_gd.comm) in ['ident', 'congruent']
        else:
            assert gd.comm.compare(aux_gd.comm) in ['ident', 'congruent']

        if aux_gd.comm.rank == 0:
            aux_ranks = gd.comm.translate_ranks(aux_gd.comm,
                                                np.arange(gd.comm.size))
        else:
            aux_ranks = np.empty(gd.comm.size, dtype=int)
        aux_gd.comm.broadcast(aux_ranks, 0)

        auxrank2rank = dict(zip(aux_ranks, np.arange(gd.comm.size)))

        def rank2parpos1(rank):
            if rank in auxrank2rank:
                return gd.get_processor_position_from_rank(auxrank2rank[rank])
            else:
                return None

        rank2parpos2 = aux_gd.get_processor_position_from_rank

        try:
            gd.n_cp
        except AttributeError:  # AtomPAW
            self._distribute = self._collect = lambda x: None
            return  # XXX

        self._distribute = partial(general_redistribute, aux_gd.comm,
                                   gd.n_cp, aux_gd.n_cp,
                                   rank2parpos1, rank2parpos2)
        self._collect = partial(general_redistribute, aux_gd.comm,
                                aux_gd.n_cp, gd.n_cp,
                                rank2parpos2, rank2parpos1)

    def distribute(self, src_xg, dst_xg=None):
        if not self.enabled:
            assert src_xg is dst_xg or dst_xg is None
            return src_xg
        if dst_xg is None:
            dst_xg = self.aux_gd.empty(src_xg.shape[:-3], dtype=src_xg.dtype)
        self._distribute(src_xg, dst_xg)
        return dst_xg

    def collect(self, src_xg, dst_xg=None):
        if not self.enabled:
            assert src_xg is dst_xg or dst_xg is None
            return src_xg
        if dst_xg is None:
            dst_xg = self.gd.empty(src_xg.shape[:-3], src_xg.dtype)
        self._collect(src_xg, dst_xg)
        self.broadcast_comm.broadcast(dst_xg, 0)
        return dst_xg

    def get_atom_distributions(self, spos_ac):
        return AtomDistributions(self.comm, self.broadcast_comm,
                                 self.gd, self.aux_gd, spos_ac)


class AtomDistributions:
    def __init__(self, comm, broadcast_comm, gd, aux_gd, spos_ac):
        self.comm = comm
        self.broadcast_comm = broadcast_comm
        self.gd = gd
        self.aux_gd = aux_gd

        rank_a = gd.get_ranks_from_positions(spos_ac)
        aux_rank_a = aux_gd.get_ranks_from_positions(spos_ac)
        self.partition = AtomPartition(gd.comm, rank_a, name='gd')

        if gd is aux_gd:
            name = 'aux-unextended'
        else:
            name = 'aux-extended'
        self.aux_partition = AtomPartition(aux_gd.comm, aux_rank_a, name=name)

        self.work_partition = AtomPartition(comm, np.zeros(len(spos_ac)),
                                            name='work').as_even_partition()

        if gd is aux_gd:
            aux_broadcast_comm = gd.comm.new_communicator([gd.comm.rank])
        else:
            aux_broadcast_comm = broadcast_comm

        self.aux_dist = AtomicMatrixDistributor(self.partition,
                                                aux_broadcast_comm,
                                                self.aux_partition)
        self.work_dist = AtomicMatrixDistributor(self.partition,
                                                 broadcast_comm,
                                                 self.work_partition)

    def to_aux(self, arraydict):
        if self.gd is self.aux_gd:
            return arraydict.copy()
        return self.aux_dist.distribute(arraydict)

    def from_aux(self, arraydict):
        if self.gd is self.aux_gd:
            return arraydict.copy()
        return self.aux_dist.collect(arraydict)

    def to_work(self, arraydict):
        return self.work_dist.distribute(arraydict)

    def from_work(self, arraydict):
        return self.work_dist.collect(arraydict)


def get_domains_from_gd(comm, gd, offset_c=None):
    ranks = gd.comm.translate_ranks(comm, np.arange(gd.comm.size))
    assert (ranks >= 0).all(), 'comm not parent of gd.comm'

    def rank2parpos(rank):
        gdrank = comm.translate_ranks(gd.comm, np.array([rank]))[0]
        # XXXXXXXXXXXXX segfault when not passing array!!
        if gdrank == -1:
            return None
        return gd.get_processor_position_from_rank(gdrank)

    def add_offset(n_cp, offset_c):
        n_cp = [n_p.copy() for n_p in n_cp]
        for c in range(3):
            n_cp[c] += offset_c[c]
        return n_cp

    n_cp = gd.n_cp
    if offset_c is not None:
        n_cp = add_offset(n_cp, offset_c)

    return n_cp, rank2parpos


def grid2grid(comm, gd1, gd2, src_g, dst_g, offset1_c=None, offset2_c=None):
    assert np.all(src_g.shape == gd1.n_c)
    assert np.all(dst_g.shape == gd2.n_c)

    n1_cp, rank2parpos1 = get_domains_from_gd(comm, gd1, offset_c=offset1_c)
    n2_cp, rank2parpos2 = get_domains_from_gd(comm, gd2, offset_c=offset2_c)

    general_redistribute(comm,
                         n1_cp, n2_cp,
                         rank2parpos1, rank2parpos2,
                         src_g, dst_g)


def main():
    from gpaw.grid_descriptor import GridDescriptor
    from gpaw.mpi import world

    serial = world.new_communicator([world.rank])

    # Generator which must run on all ranks
    gen = np.random.RandomState(0)

    # This one is just used by master
    gen_serial = np.random.RandomState(17)

    maxsize = 5
    for i in range(1):
        N1_c = gen.randint(1, maxsize, 3)
        N2_c = gen.randint(1, maxsize, 3)

        gd1 = GridDescriptor(N1_c, N1_c)
        gd2 = GridDescriptor(N2_c, N2_c)
        serial_gd1 = gd1.new_descriptor(comm=serial)
        # serial_gd2 = gd2.new_descriptor(comm=serial)

        a1_serial = serial_gd1.empty()
        a1_serial.flat[:] = gen_serial.rand(a1_serial.size)

        if world.rank == 0:
            print('r0: a1 serial', a1_serial.ravel())

        a1 = gd1.empty()
        a1[:] = -1

        grid2grid(world, serial_gd1, gd1, a1_serial, a1)

        print(world.rank, 'a1 distributed', a1.ravel())
        world.barrier()

        a2 = gd2.zeros()
        a2[:] = -2
        grid2grid(world, gd1, gd2, a1, a2)
        print(world.rank, 'a2 distributed', a2.ravel())
        world.barrier()

        gd1 = GridDescriptor(N1_c, N1_c * 0.2)

        a1 = gd1.empty()
        a1.flat[:] = gen.rand(a1.size)

        grid2grid(world, gd1, gd2, a1, a2)


if __name__ == '__main__':
    main()



"""
Calculates Raman matrices and intensities.

i -> j -> m -> n
i, n are valence; j, m are conduction, also i=n in the end
see https://doi.org/10.1038/s41467-020-16529-6
"""

import numpy as np
from ase.units import invcm, Hartree


def lorentzian(w, gamma):
    l = 0.5 * gamma / (np.pi * (w**2 + 0.25 * gamma**2))
    return l


def gaussian(w, sigma):
    g = 1. / (np.sqrt(2. * np.pi) * sigma) * np.exp(-w**2 / (2 * sigma**2))
    return g


def calculate_raman(calc, w_ph, w_in, d_i, d_o, resonant_only=False,
                    gamma_l=0.1, momname='mom_skvnm.npy',
                    elphname='gsqklnn.npy', gridspacing=1.0, suffix=None):
    """
    Calculates the first order Raman tensor contribution for a given
    polarization.

    Parameters
    ----------
    calc: GPAW
        Converged ground state calculation
    w_ph: np.ndarray, str
        Array of phonon frequencies in eV, or name of file with them
    w_in: float
        Laser energy in eV
    d_i: int
        Incoming polarization
    d_o: int
        Outgoing polarization
    gamma_l: float
        Line broadening in eV, (0.1eV=806rcm)
    momname: str
        Name of momentum file
    elphname: str
        Name of electron-phonon file
    gridspacing: float
        grid spacing in cm^-1
    suffix: str
        suffix for Rlab file name. Defaults to XXXnm if not given.
    """
    # those won't make sense here
    assert calc.wfs.gd.comm.size == 1
    assert calc.wfs.bd.comm.size == 1
    kd = calc.wfs.kd

    print("Calculating Raman spectrum: Laser frequency = {} eV".format(w_in))
    if suffix is None:
        suffix = "{}nm".format(int(1239.841 / w_in))

    # Phonon frequencies
    if isinstance(w_ph, str):
        w_ph = np.load(w_ph)
    assert max(w_ph) < 1.  # else not eV units
    nmodes = len(w_ph)

    # Set grid
    w_max = np.round(np.max(w_ph) / invcm + 50, -1)  # max of grid in rcm
    ngrid = int(w_max / gridspacing) + 1
    w = np.linspace(0., w_max, num=ngrid) * invcm  # in eV

    # Load files
    mom_skvnm = np.load(momname, mmap_mode='c')
    g_sqklnn = np.load(elphname, mmap_mode='c')  # [s,q=0,k,l,n,m]

    # Define a few more variables
    opt = 'optimal'  # mode for np.einsum. not sure what is fastest
    nspins = g_sqklnn.shape[0]
    nk = g_sqklnn.shape[2]
    assert nmodes == g_sqklnn.shape[3]
    assert mom_skvnm.shape[0] == nspins
    assert mom_skvnm.shape[1] == nk
    assert mom_skvnm.shape[-1] == g_sqklnn.shape[-1]

    # valence is ket in ij and bra in mn
    # ab is in and out polarization
    # l is the phonon mode and w is the raman shift

    # XXX: The below can probably be made better by lambdas a lot
    def _term1(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term1_l = np.zeros((elph_lnn.shape[0]), dtype=complex)
        t1_ij = f_vc * mom_dnn[0, :nv, nc:] / (w_in - E_vc)
        # t1_ij = f_vc * mom_dnn[0, nc:, :nv].T / (w_in - E_vc)
        for l in range(nmodes):
            t1_xx = elph_lnn[l]
            # t1_mn = (f_vc * mom_dnn[1, :nv, nc:] / (w_in - w_ph[l] - E_vc)).T
            t1_mn = f_vc.T * mom_dnn[1, nc:, :nv] / (w_in - w_ph[l] - E_vc.T)
            term1_l[l] += np.einsum('sj,jm,ms', t1_ij, t1_xx[nc:, nc:], t1_mn,
                                    optimize=opt)
            term1_l[l] -= np.einsum('is,ni,sn', t1_ij, t1_xx[:nv, :nv], t1_mn,
                                    optimize=opt)
        return term1_l

    def _term2(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term2_lw = np.zeros((nmodes, ngrid), dtype=complex)
        t2_ij = f_vc * mom_dnn[0, :nv, nc:] / (w_in - E_vc)
        t2_xx = mom_dnn[1]
        for l in range(nmodes):
            for wi in range(ngrid):
                t2_mn = f_vc.T * elph_lnn[l][nc:, :nv] / (w[wi] - E_vc.T)
                term2_lw[l, wi] += np.einsum('sj,jm,ms', t2_ij,
                                             t2_xx[nc:, nc:], t2_mn,
                                             optimize=opt)
                term2_lw[l, wi] -= np.einsum('is,ni,sn', t2_ij,
                                             t2_xx[:nv, :nv], t2_mn,
                                             optimize=opt)
        return term2_lw

    def _term3(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term3_lw = np.zeros((nmodes, ngrid), dtype=complex)
        for l in range(nmodes):
            t3_xx = elph_lnn[l]
            for wi in range(ngrid):
                t3_ij = f_vc * mom_dnn[1, :nv, nc:] / (-w_in + w[wi] - E_vc)
                t3_mn = (f_vc.T * mom_dnn[0, :nv, nc:] / (-w_in - w_ph[l] +
                                                          w[wi] - E_vc.T))
                term3_lw[l, wi] += np.einsum('sj,jm,ms', t3_ij,
                                             t3_xx[nc:, nc:], t3_mn,
                                             optimize=opt)
                term3_lw[l, wi] -= np.einsum('is,ni,sn', t3_ij,
                                             t3_xx[:nv, :nv], t3_mn,
                                             optimize=opt)
        return term3_lw

    def _term4(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term4_lw = np.zeros((nmodes, ngrid), dtype=complex)
        t4_xx = mom_dnn[0]
        for l in range(nmodes):
            for wi in range(ngrid):
                t4_ij = f_vc * mom_dnn[1, :nv, nc:] / (-w_in + w[wi] - E_vc)
                t4_mn = (f_vc.T * elph_lnn[l, nc:, :nv] / (w[wi] - E_vc.T))
                term4_lw[l, wi] += np.einsum('sj,jm,ms', t4_ij,
                                             t4_xx[nc:, nc:], t4_mn,
                                             optimize=opt)
                term4_lw[l, wi] -= np.einsum('is,ni,sn', t4_ij,
                                             t4_xx[:nv, :nv], t4_mn,
                                             optimize=opt)
        return term4_lw

    def _term5(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term5_l = np.zeros((nmodes), dtype=complex)
        t5_xx = mom_dnn[0]
        for l in range(nmodes):
            t5_ij = f_vc * elph_lnn[l, :nv, nc:] / (-w_ph[l] - E_vc)
            t5_mn = (f_vc.T * mom_dnn[1, nc:, :nv] / (w_in - w_ph[l] - E_vc.T))
            term5_l[l] += np.einsum('sj,jm,ms', t5_ij, t5_xx[nc:, nc:], t5_mn,
                                    optimize=opt)
            term5_l[l] -= np.einsum('is,ni,sn', t5_ij, t5_xx[:nv, :nv], t5_mn,
                                    optimize=opt)
        return term5_l

    def _term6(f_vc, E_vc, mom_dnn, elph_lnn, nc, nv):
        term6_lw = np.zeros((nmodes, ngrid), dtype=complex)
        t6_xx = mom_dnn[1]
        for l in range(nmodes):
            t6_ij = f_vc * elph_lnn[l, :nv, nc:] / (-w_ph[l] - E_vc)
            for wi in range(ngrid):
                t6_mn = (f_vc.T * mom_dnn[0, nc:, :nv] / (-w_in - w_ph[l]
                                                          + w[wi] - E_vc.T))
                term6_lw[l, wi] += np.einsum('sj,jm,ms', t6_ij,
                                             t6_xx[nc:, nc:], t6_mn,
                                             optimize=opt)
                term6_lw[l, wi] -= np.einsum('is,ni,sn', t6_ij,
                                             t6_xx[:nv, :nv], t6_mn,
                                             optimize=opt)
        return term6_lw

    # -------------------------------------------------------------------------

    print("Evaluating Raman sum")
    raman_lw = np.zeros((nmodes, ngrid), dtype=complex)

    # Loop over kpoints - this is parallelised
    for kpt in calc.wfs.kpt_u:
        print("Rank {}: s={}, k={}".format(kd.comm.rank, kpt.s, kpt.k))

        # Check if we need to add timer-add time reversed kpoint
        if (calc.symmetry.time_reversal and not
            np.allclose(kd.ibzk_kc[kpt.k], [0., 0., 0.])):
            add_time_reversed = True
            # Currently broken
            raise NotImplementedError
        else:
            add_time_reversed = False

        # Limit sums to relevant bands, partially occupied bands are a pain.
        # So, in principal, partially occupied bands can be initial and
        # final states, but we need to restrict to a positive deltaE if we
        # allow this.
        f_n = kpt.f_n / kpt.weight
        assert np.isclose(max(f_n), 1.0, atol=0.1)

        if 1:
            vs = np.arange(0, len(f_n))
            cs = np.arange(0, len(f_n))
            nc = 0
            nv = len(f_n)
        else:
            vs = np.where(f_n >= 0.1)[0]
            cs = np.where(f_n < 0.9)[0]
            nv = max(vs) + 1  # VBM+1 index
            nc = min(cs)  # CBM index

        # Precalculate f * (1-f) term
        f_vc = np.outer(f_n[vs], 1. - f_n[cs])

        # Precalculate E-E term
        E_vc = np.empty((len(vs), len(cs)), dtype=complex)
        for n in range(len(vs)):
            E_vc[n] = (kpt.eps_n[cs] - kpt.eps_n[n]) * Hartree + 1j * gamma_l

            # set weights for negative energy transitions zero
            if 0:
                neg = np.where(E_vc[n].real <= 0.)[0]
                f_vc[n, neg] = 0.

        # Obtain appropriate part of mom and g arrays
        mom_dnn = np.ascontiguousarray(mom_skvnm[kpt.s, kpt.k, [d_i, d_o]])
        assert mom_dnn.shape[0] == 2
        g_lnn = np.ascontiguousarray(g_sqklnn[kpt.s, 0, kpt.k])

        # Raman contribution of this k-point
        this_lw = np.zeros((nmodes, ngrid), dtype=complex)

        # Resonant term
        term1_l = _term1(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
        # print("Term1: ", np.max(np.abs(term1_l)))
        this_lw += term1_l[:, None]

        if not resonant_only:
            term2_lw = _term2(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
            # print("Term2: ", np.max(np.abs(term2_lw)))
            this_lw += term2_lw

            term3_lw = _term3(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
            # print("Term3: ", np.max(np.abs(term3_lw)))
            this_lw += term3_lw

            term4_lw = _term4(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
            # print("Term4: ", np.max(np.abs(term4_lw)))
            this_lw += term4_lw

            term5_l = _term5(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
            # print("Term5: ", np.max(np.abs(term5_l)))
            this_lw += term5_l[:, None]

            term6_lw = _term6(f_vc, E_vc, mom_dnn, g_lnn, nc, nv)
            # print("Term6: ", np.max(np.abs(term6_lw)))
            this_lw += term6_lw

        # At the moment we only allow no symmetry, so all k-points same
        # weight, or time_reversal only. In the later case r -> 2*Re(r)
        # because gdd-> (gdd)^* for k-> -k
        if add_time_reversed:
            raman_lw += 2. * this_lw.real
        else:
            raman_lw += this_lw

    # Collect parallel contributions
    kd.comm.sum(raman_lw)

    if kd.comm.rank == 0:
        print('Raman intensities per mode')
        print('--------------------------')
        ff = '  Phonon {} with energy = {:4.2f} rcm: {:.4f}'
        for l in range(nmodes):
            print(ff.format(l, w_ph[l] / invcm, abs(np.sum(raman_lw[l]))))

        # Save Raman intensities to disk
        raman = np.vstack((w / invcm, raman_lw))
        np.save("Rlab_{}{}_{}.npy".format('xyz'[d_i], 'xyz'[d_o], suffix),
                raman)


def calculate_raman_tensor(calc, w_ph, w_in, resonant_only=False,
                           gamma_l=0.1, momname='mom_skvnm.npy',
                           elphname='gsqklnn.npy'):
    for i in range(3):
        for j in range(3):
            calculate_raman(calc, w_ph, w_in, d_i=i, d_o=j,
                            resonant_only=resonant_only, gamma_l=gamma_l,
                            momname=momname, elphname=elphname)


def calculate_raman_intensity(w_ph, d_i, d_o, suffix, T=300):
    """
    Calculates Raman intensities from Raman tensor.

    Parameters
    ----------
    w_ph: np.ndarray, str
        Array of phonon frequencies in eV, or name of file with them
    d_i: int
        Incoming polarization
    d_o: int
        Outgoing polarization
    suffix: str
        Suffix for Rlab and RI files. Usually XXXnm
    """
    KtoeV = 8.617278e-5
    # Phonon frequencies
    if isinstance(w_ph, str):
        w_ph = np.load(w_ph)
    assert max(w_ph) < 1.  # else not eV units

    # Load raman matrix elements R_lab
    tmp = np.load("Rlab_{}{}_{}.npy".format('xyz'[d_i], 'xyz'[d_o], suffix))
    w = tmp[0].real  # already in rcm
    raman_lw = tmp[1:]
    assert raman_lw.shape[0] == len(w_ph)

    int_bare = np.zeros_like(w, dtype=float)
    int_occ = np.zeros_like(w, dtype=float)
    for l in range(len(w_ph)):
        occ = 1. / (np.exp(w_ph[l] / (KtoeV * T)) - 1.) + 1.
        delta = gaussian(w=(w - w_ph[l] / invcm), sigma=5.)
        R = raman_lw[l] * raman_lw[l].conj()
        int_bare += R.real * delta
        int_occ += occ / w_ph[l] * R.real * delta

    raman = np.vstack((w, int_bare, int_occ))
    np.save("RI_{}{}_{}.npy".format('xyz'[d_i], 'xyz'[d_o], suffix), raman)


def plot_raman(figname, RIsuffix, relative=False, w_min=None, w_max=None):
    """Plots a given Raman spectrum.

    Parameters
    ----------
    figname: str
        Filename for figure.
    RIsuffix: str, list
        Suffix of Raman intensity files to use for plotting. For example
        "0_1_455nm" for RI_0_1_455nm.npy
    """
    from scipy import signal
    import matplotlib.pyplot as plt
    import matplotlib.colors as colors
    import matplotlib.cm as cmx

    if isinstance(RIsuffix, str):
        legend = False
        RI_name = ["RI_{}.npy".format(RIsuffix)]
    else:  # assume list
        legend = True
        RI_name = ["RI_{}.npy".format(name) for name in RIsuffix]
        cm = plt.get_cmap('inferno')
        cNorm = colors.Normalize(vmin=0, vmax=len(RI_name))
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cm)

    ylabel = "Intensity (arb. units)"
    if w_min is None:
        w_min = 0.

    peaks = None
    for i, name in enumerate(RI_name):
        RI = np.load(name)  # load raman intensity
        if w_max is None:
            w_max = np.max(RI[0])
        r = RI[1][np.logical_and(RI[0] >= w_min, RI[0] <= w_max)]
        w = RI[0][np.logical_and(RI[0] >= w_min, RI[0] <= w_max)]
        if relative:
            ylabel = "I/I_max"
            r /= np.max(r)

        if peaks is None:
            peaks = signal.find_peaks(r[np.logical_and(w >= w_min, w <= w_max)
                                        ])[0]
            locations = np.take(w[np.logical_and(w >= w_min, w <= w_max)],
                                peaks)
            intensities = np.take(r[np.logical_and(w >= w_min, w <= w_max)],
                                  peaks)

        if legend:
            cval = scalarMap.to_rgba(i)
            plt.plot(w, r, color=cval, label=RIsuffix[i])
        else:
            plt.plot(w, r)

    for i, loc in enumerate(locations):
        if intensities[i] / np.max(intensities) > 0.05:
            plt.axvline(x=loc, color="grey", linestyle="--")

    plt.minorticks_on()
    if legend:
        plt.legend()
    plt.title("Raman intensity")
    plt.xlabel("Raman shift (cm$^{-1}$)")
    plt.ylabel(ylabel)
    if relative:
        plt.yticks([])
    plt.savefig(figname, dpi=300)
    plt.clf()



import numpy as np
import ase.units
from scipy.linalg import eigh

from gpaw.mpi import world
from gpaw.lrtddft2.lr_layouts import LrDiagonalizeLayout


class LrtddftTransitions:
    def __init__(self, ks_singles, K_matrix, sl_lrtddft=None):
        self.ks_singles = ks_singles
        self.K_matrix = K_matrix
        self.lr_comms = self.ks_singles.lr_comms
        self.sl_lrtddft = sl_lrtddft
        self.eigenvalues = None
        self.eigenvectors = None
        self.trans_prop_ready = False

        # only for pros
        self.custom_axes = None

    def initialize(self):
        self.trans_prop_ready = False

    def read(self):
        pass

    def calculate(self):
        self.diagonalize()
        self.calculate_properties()

    def diagonalize(self):
        if self.sl_lrtddft is None:
            self.diagonalize_serial()
        else:
            self.diagonalize_scalapack()

        # now self.eigenvectors[iploc,kloc]
        # is the "iploc"th local element of
        # the "kloc"th local eigenvector

    def diagonalize_serial(self):
        nrows = len(self.ks_singles.kss_list)

        # build local part of the full omega matrix
        omega_matrix = np.zeros((nrows, nrows), dtype=float)
        for (ip, kss_ip) in enumerate(self.ks_singles.kss_list):
            for (jq, kss_jq) in enumerate(self.ks_singles.kss_list):
                lip = self.lr_comms.get_local_eh_index(ip)
                ljq = self.lr_comms.get_local_dd_index(jq)
                if lip is None or ljq is None:
                    continue

                # K-matrix
                # if self.K_matrix.file_format == 1:
                omega_matrix[ip, jq] = 2. * np.sqrt(
                    kss_ip.energy_diff * kss_jq.energy_diff * kss_ip.pop_diff *
                    kss_jq.pop_diff) * self.K_matrix.values[lip, ljq]
                # old Casida format ... doing this already in k_matrix.py
                # elif self.K_matrix.file_format == 0:
                #    omega_matrix[ip,jq] = self.K_matrix.values[lip,ljq]
                # invalid
                # else:
                #    raise RuntimeError('Invalid K-matrix file format')

                if (ip == jq):
                    omega_matrix[ip,
                                 jq] += kss_ip.energy_diff * kss_jq.energy_diff

        # full omega matrix to each process
        self.lr_comms.parent_comm.sum(omega_matrix)

        # if self.lr_comms.parent_comm.rank == 0:
        #    print omega_matrix

        # solve eigensystem

        # debug
        # omega_matrix[:] = 0
        # for i in range(omega_matrix.shape[0]):
        #    for j in range(omega_matrix.shape[1]):
        #        if i == j: omega_matrix[i,j] = -2.
        #        if i+1 == j: omega_matrix[i,j] = omega_matrix[j,i] = 1.

        self.eigenvalues, omega_matrix = eigh(omega_matrix)

        # convert eigenvector back to local version
        nlrows = self.K_matrix.values.shape[0]
        nlcols = self.K_matrix.values.shape[1]

        # debug
        # (eh=2,dd=3)      (0,0)  (0,1)  (0,2)  (1,0)  (1,1)  (1,2)
        # 00 01 02 03      00 03    01     02   10 13    11     12
        # 10 11 12 13   => 20 23    21     22   30 33    31     32
        # 20 21 22 23
        # 30 31 32 33

        self.eigenvectors = np.zeros((nlrows, nlcols), dtype=float)
        for (ip, kss) in enumerate(self.ks_singles.kss_list):
            for (jq, kss) in enumerate(self.ks_singles.kss_list):
                lip = self.lr_comms.get_local_eh_index(ip)
                ljq = self.lr_comms.get_local_dd_index(jq)
                if lip is None or ljq is None:
                    continue
                self.eigenvectors[lip,
                                  ljq] = omega_matrix[ip,
                                                      jq]  # debug: ip*10 + jq

    def diagonalize_scalapack(self):
        # total rows
        nrows = len(self.ks_singles.kss_list)
        # local
        nlrows = self.K_matrix.values.shape[0]
        nlcols = self.K_matrix.values.shape[1]

        # create BLACS layout
        layout = LrDiagonalizeLayout(self.sl_lrtddft, nrows, self.lr_comms)
        if (nrows <
                np.max([layout.mprocs, layout.nprocs]) * layout.block_size):
            raise RuntimeError(
                'Linear response matrix is not large enough for the given '
                'number of processes (or block size) in sl_lrtddft. Please, '
                'use less processes (or smaller block size).'
            )

        # build local part
        omega_matrix = np.zeros((nlrows, nlcols), dtype=float)
        for (ip, kss_ip) in enumerate(self.ks_singles.kss_list):
            for (jq, kss_jq) in enumerate(self.ks_singles.kss_list):
                lip = self.lr_comms.get_local_eh_index(ip)
                ljq = self.lr_comms.get_local_dd_index(jq)
                if lip is None or ljq is None:
                    continue

                # if self.K_matrix.file_format == 1:
                omega_matrix[lip, ljq] = 2. * np.sqrt(
                    kss_ip.energy_diff * kss_jq.energy_diff * kss_ip.pop_diff *
                    kss_jq.pop_diff) * self.K_matrix.values[lip, ljq]
                # old Casida format ... doing this already in k_matrix.py
                # elif self.K_matrix.file_format == 0:
                #    omega_matrix[lip,ljq] = self.K_matrix.values[lip,ljq]
                # else:
                #    raise RuntimeError('Invalid K-matrix file format')

                if (ip == jq):
                    omega_matrix[
                        lip, ljq] += kss_ip.energy_diff * kss_jq.energy_diff

        # solve eigen system (transpose for scalapack)
        self.eigenvalues = np.zeros(nrows, dtype=float)
        self.eigenvectors = np.zeros(
            (omega_matrix.shape[1], omega_matrix.shape[0]), dtype=float)
        self.eigenvectors[:, :] = np.transpose(omega_matrix)
        layout.diagonalize(self.eigenvectors, self.eigenvalues)
        omega_matrix[:, :] = np.transpose(self.eigenvectors)
        self.eigenvectors = omega_matrix

    def calculate_properties(self):
        if self.custom_axes is not None:
            self.custom_axes = np.array(self.custom_axes)
        # else:
        #    self.custom_axes = np.array([[1.0,0.0,0.0],
        #                                 [0.0,1.0,0.0],
        #                                 [0.0,0.0,1.0]])

        self.lr_comms.parent_comm.barrier()

        # nlrows = self.eigenvectors.shape[0]
        nleigs = self.eigenvectors.shape[1]

        sqrtwloc = np.zeros(nleigs)
        dmxloc = np.zeros(nleigs)
        dmyloc = np.zeros(nleigs)
        dmzloc = np.zeros(nleigs)
        magnxloc = np.zeros(nleigs)
        magnyloc = np.zeros(nleigs)
        magnzloc = np.zeros(nleigs)
        cloc_dm = np.zeros(nleigs)
        cloc_magn = np.zeros(nleigs)

        # see Autschbach et al., J. Chem. Phys., 116, 6930 (2002)
        # see also Varsano et al., Phys.Chem.Chem.Phys. 11, 4481 (2009)
        #
        # mu_k = sum_jq sqrt(ediff_jq / omega_k)
        #        sqrt(population_jq) * F^(k)_jq * mu_jq
        # m_k  = sum_jq sqrt(omega_k  / ediff_jq)
        #        sqrt(population_jq) * F^(k)_jq * m_jq
        #
        # FIXME: check once more
        # mu_k = sum_jq c1_k * c2_k * mu_jq
        #      = sum_jq sqrt(ediff_jq / omega_k)
        #        sqrt(fdiff_jq) * F^(k)_jq mu_jq
        # m_k  = sum_jq c2_k / c1_k * m_jq
        #      = sum_jq sqrt(fdiff_jq) sqrt(omega_k / ediff_jq) * F^(k)_jq m_jq
        #
        # c1 =  sqrt(ediff_jq / omega_k)
        #    =  np.sqrt(kss_jq.energy_diff / self.get_excitation_energy(k))
        #
        # c2 = sqrt(fdiff_jq) * F^(k)_jq
        #    = np.sqrt(kss_jq.pop_diff) *
        #      self.get_local_eig_coeff(k,self.index_map[(i,p)])
        #

        # sqrt(omega_kloc)
        for kloc in range(nleigs):
            sqrtwloc[kloc] = np.sqrt(
                np.sqrt(
                    self.eigenvalues[self.lr_comms.get_global_dd_index(kloc)]))

        # loop over ks singles
        for (ip, kss_ip) in enumerate(self.ks_singles.kss_list):

            # local ip index
            lip = self.lr_comms.get_local_eh_index(ip)
            # print self.lr_comms.parent_comm.rank, i,p,ip,lip
            if lip is None:
                continue

            # now self.eigenvectors[iploc,kloc]
            # is the "iploc"th local element of
            # the "kloc"th local eigenvector
            #
            # init local mu and m sums with
            # a continuous copy of ROW of eigenvectors
            # i.e. a "ip"th element of each eigenvector
            # (NOT the "jq"th eigenvector)
            cloc_dm[:] = self.eigenvectors[lip, :]
            cloc_magn[:] = self.eigenvectors[lip, :]

            # c1 * c2 * F = sqrt(fdiff_ip * ediff_ip) / sqrt(omega_k) F^(k)_ip
            cloc_dm *= np.sqrt(kss_ip.pop_diff * kss_ip.energy_diff)
            cloc_dm /= sqrtwloc

            # c2 / c1 * F = sqrt(fdiff_ip / ediff_ip) * sqrt(omega_k) F^(k)_ip
            cloc_magn *= np.sqrt(kss_ip.pop_diff / kss_ip.energy_diff)
            cloc_magn *= sqrtwloc

            if self.custom_axes is None:
                dmxloc += kss_ip.dip_mom_r[0] * cloc_dm
                dmyloc += kss_ip.dip_mom_r[1] * cloc_dm
                dmzloc += kss_ip.dip_mom_r[2] * cloc_dm

                magnxloc += kss_ip.magn_mom[0] * cloc_magn
                magnyloc += kss_ip.magn_mom[1] * cloc_magn
                magnzloc += kss_ip.magn_mom[2] * cloc_magn
            else:
                dmxloc += np.dot(kss_ip.dip_mom_r,
                                 self.custom_axes[0]) * cloc_dm
                dmyloc += np.dot(kss_ip.dip_mom_r,
                                 self.custom_axes[1]) * cloc_dm
                dmzloc += np.dot(kss_ip.dip_mom_r,
                                 self.custom_axes[2]) * cloc_dm

                magnxloc += np.dot(kss_ip.magn_mom,
                                   self.custom_axes[0]) * cloc_magn
                magnyloc += np.dot(kss_ip.magn_mom,
                                   self.custom_axes[1]) * cloc_magn
                magnzloc += np.dot(kss_ip.magn_mom,
                                   self.custom_axes[2]) * cloc_magn

        # global properties, but local eigs
        # sum over iploc dmx_iploc[kloc] to dmx[kloc]
        props = np.array(
            [dmxloc, dmyloc, dmzloc, magnxloc, magnyloc, magnzloc])
        self.lr_comms.eh_comm.sum(props.ravel())
        [dmxloc, dmyloc, dmzloc, magnxloc, magnyloc, magnzloc] = props

        # global properties and eigs
        # create local part of the global eigs
        nrows = len(self.ks_singles.kss_list)
        self.transition_properties = np.zeros([nrows, 1 + 3 + 3])
        for k in range(nrows):
            kloc = self.lr_comms.get_local_dd_index(k)
            if kloc is None:
                continue
            self.transition_properties[k, 0] = sqrtwloc[kloc] * sqrtwloc[kloc]
            self.transition_properties[k, 1] = dmxloc[kloc]
            self.transition_properties[k, 2] = dmyloc[kloc]
            self.transition_properties[k, 3] = dmzloc[kloc]
            self.transition_properties[k, 4] = magnxloc[kloc]
            self.transition_properties[k, 5] = magnyloc[kloc]
            self.transition_properties[k, 6] = magnzloc[kloc]

        # sum over kloc dmx[...kloc...] to dmx[k]
        self.lr_comms.dd_comm.sum(self.transition_properties.ravel())

        # print self.lr_comms.parent_comm.rank, self.transition_properties[:,1]
        # return

        self.trans_prop_ready = True

    # omega_k = sqrt(lambda_k)
    def get_excitation_energy(self, k, units='au'):
        """Get excitation energy for kth interacting transition

        Input parameters:

        k
          Transition index (indexing starts from zero).

        units
          Units for excitation energy: 'au' (Hartree) or 'eV'.
        """
        if not self.trans_prop_ready:
            self.calculate()

        if units == 'au':
            return np.sqrt(self.eigenvalues[k])
        elif units == 'eV' or units == 'eVcgs':
            return np.sqrt(self.eigenvalues[k]) * ase.units.Hartree
        else:
            raise RuntimeError('Unknown units.')

    def get_oscillator_strength(self, k):
        """Get oscillator strength for an interacting transition

        Returns oscillator strength of kth interacting transition.

        Input parameters:

        k
          Transition index (indexing starts from zero).
        """
        if not self.trans_prop_ready:
            self.calculate()

        omega = self.transition_properties[k][0]
        dmx = self.transition_properties[k][1]
        dmy = self.transition_properties[k][2]
        dmz = self.transition_properties[k][3]

        oscx = 2. * omega * dmx * dmx
        oscy = 2. * omega * dmy * dmy
        oscz = 2. * omega * dmz * dmz
        osc = (oscx + oscy + oscz) / 3.
        return osc, np.array([oscx, oscy, oscz])

    def get_rotatory_strength(self, k, units='au'):
        """Get rotatory strength for an interacting transition

        Returns rotatory strength of kth interacting transition.

        Input parameters:

        k
          Transition index (indexing starts from zero).

        units
          Units for rotatory strength: 'au' or 'cgs'.
        """

        if not self.trans_prop_ready:
            self.calculate()

        # omega = self.transition_properties[k][0]
        dmx = self.transition_properties[k][1]
        dmy = self.transition_properties[k][2]
        dmz = self.transition_properties[k][3]
        magnx = self.transition_properties[k][4]
        magny = self.transition_properties[k][5]
        magnz = self.transition_properties[k][6]

        if units == 'au':
            return -(dmx * magnx + dmy * magny + dmz * magnz)
        elif units == 'cgs' or units == 'eVcgs':
            # 10^-40 esu cm erg / G
            # = 3.33564095 * 10^-15 A^2 m^3 s
            # conversion factor 471.43 from
            # T. B. Pedersen and A. E. Hansen,
            # Chem. Phys. Lett. 246 (1995) 1
            # From turbomole 64604.8164
            #
            # FIX ME: why?
            # 64604.8164/471.43 = 137.040
            # is the inverse of fine-structure constant
            # OR it must be speed of light in atomic units
            return -64604.8164 * (dmx * magnx + dmy * magny + dmz * magnz)
        else:
            raise RuntimeError('Unknown units.')

    def get_transitions(self,
                        filename=None,
                        min_energy=0.0,
                        max_energy=30.0,
                        units='eVcgs'):
        """Get transitions: energy, dipole strength and rotatory strength.

        Returns transitions as (w,S,R, Sx,Sy,Sz) where w is an array of
        frequencies,
        S is an array of corresponding dipole strengths, and R is an array of
        corresponding rotatory strengths.

        Input parameters:

        min_energy
          Minimum energy

        min_energy
          Maximum energy

        units
          Units for spectrum: 'au' or 'eVcgs'
        """
        if not self.trans_prop_ready:
            self.calculate()

        w = []
        S = []
        R = []
        Sx = []
        Sy = []
        Sz = []

        for (k, omega) in enumerate(self.eigenvalues):
            ww = self.get_excitation_energy(k, units)
            if ww < min_energy or ww > max_energy:
                continue

            w.append(ww)
            St, Sc = self.get_oscillator_strength(k)
            S.append(St)
            Sx.append(Sc[0])
            Sy.append(Sc[1])
            Sz.append(Sc[2])
            R.append(self.get_rotatory_strength(k, units))

        w = np.array(w)
        S = np.array(S)
        R = np.array(R)
        Sx = np.array(Sx)
        Sy = np.array(Sy)
        Sz = np.array(Sz)

        if filename is not None and world.rank == 0:
            sfile = open(filename, 'w')
            sfile.write("# %12s  %12s  %12s     %12s  %12s  %12s    %s\n" %
                        ('energy', 'osc str', 'rot str', 'osc str x',
                         'osc str y', 'osc str z', 'units: ' + units))
            for (ww, SS, RR, SSx, SSy, SSz) in zip(w, S, R, Sx, Sy, Sz):
                sfile.write(
                    ("  %12.8lf  %12.8lf  %12.8lf     "
                     "%12.8lf  %12.8lf  %12.8lf\n")
                    % (ww, SS, RR, SSx, SSy, SSz))
            sfile.close()

        return (w, S, R, Sx, Sy, Sz)

    def get_spectrum(self,
                     filename=None,
                     min_energy=0.0,
                     max_energy=30.0,
                     energy_step=0.01,
                     width=0.1,
                     units='eVcgs'):
        """Get spectrum for dipole and rotatory strength.

        Returns folded spectrum as (w,S,R) where w is an array of frequencies,
        S is an array of corresponding dipole strengths, and R is an array of
        corresponding rotatory strengths.

        Input parameters:

        min_energy
          Minimum energy

        min_energy
          Maximum energy

        energy_step
          Spacing between calculated energies

        width
          Width of the Gaussian

        units
          Units for spectrum: 'au' or 'eVcgs'
        """

        if not self.trans_prop_ready:
            self.calculate()

        (ww, SS, RR, SSx, SSy,
         SSz) = self.get_transitions(min_energy=min_energy - 5 * width,
                                     max_energy=max_energy + 5 * width,
                                     units=units)

        if units == 'eVcgs':
            pass  # convf = 1/ase.units.Hartree
        elif units == 'au':
            assert 0  # convf = 1.
        else:
            raise RuntimeError('Invalid units')

        w = np.arange(min_energy, max_energy, energy_step)
        S = np.zeros_like(w)
        Sx = np.zeros_like(w)
        Sy = np.zeros_like(w)
        Sz = np.zeros_like(w)
        R = np.zeros_like(w)

        for (k, www) in enumerate(ww):

            c = SS[k] / width / np.sqrt(2 * np.pi)
            S += c * np.exp((-.5 / width / width) * np.power(w - ww[k], 2))

            c = SSx[k] / width / np.sqrt(2 * np.pi)
            Sx += c * np.exp((-.5 / width / width) * np.power(w - ww[k], 2))

            c = SSy[k] / width / np.sqrt(2 * np.pi)
            Sy += c * np.exp((-.5 / width / width) * np.power(w - ww[k], 2))

            c = SSz[k] / width / np.sqrt(2 * np.pi)
            Sz += c * np.exp((-.5 / width / width) * np.power(w - ww[k], 2))

            c = RR[k] / width / np.sqrt(2 * np.pi)
            R += c * np.exp((-.5 / width / width) * np.power(w - ww[k], 2))

        if filename is not None and world.rank == 0:
            sfile = open(filename, 'w')
            sfile.write("# %12s  %12s  %12s     %12s  %12s  %12s    %s\n" %
                        ('energy', 'osc str', 'rot str', 'osc str x',
                         'osc str y', 'osc str z', 'units: ' + units))
            for (ww, SS, RR, SSx, SSy, SSz) in zip(w, S, R, Sx, Sy, Sz):
                sfile.write(
                    ("  %12.8lf  %12.8lf  %12.8lf     "
                     "%12.8lf  %12.8lf  %12.8lf\n")
                    % (ww, SS, RR, SSx, SSy, SSz))
            sfile.close()

        return (w, S, R, Sx, Sy, Sz)

    def get_transition_contributions(self, index_of_transition):
        """Get contributions of Kohn-Sham singles to a given transition.

        Includes population difference.

        For large systems this is slow.

        Input parameters:

        index_of_transition:
          index of transition starting from zero
        """

        neigs = len(self.ks_singles.kss_list)
        f2 = np.zeros(neigs)

        if index_of_transition < 0:
            raise RuntimeError(
                'Error in get_transition_contributions: Index < zero.')
        if index_of_transition >= neigs:
            raise RuntimeError(
                'Error in get_transition_contributions: '
                'Index >= number of transitions'
            )

        k = index_of_transition
        # local k index
        kloc = self.lr_comms.get_local_dd_index(k)

        if kloc is not None:

            for (ip, kss_ip) in enumerate(self.ks_singles.kss_list):
                # local ip index
                lip = self.lr_comms.get_local_eh_index(ip)
                if lip is None:
                    continue

                # self.eigenvectors[iploc,kloc]
                # is the "iploc"th local element of
                # the "kloc"th local eigenvector
                f2[ip] = self.eigenvectors[lip, kloc] * self.eigenvectors[
                    lip, kloc] * kss_ip.pop_diff

        self.lr_comms.parent_comm.sum(f2)

        return f2



import numpy as np
import pytest

from gpaw.atom.generator2 import generate
from gpaw.setup import Setup
from gpaw.xc import XC


@pytest.mark.serial
def test_lithium(in_tmp_dir):
    G = generate('Li', 3, '2s,2p,s', [2.1, 2.1], 2.0, None, 2, 'PBE', True)
    assert G.check_all()
    basis = G.create_basis_set()
    basis.write_xml()
    setup = G.make_paw_setup('test')
    setup.write_xml()


@pytest.mark.serial
def test_pseudo_h(in_tmp_dir):
    G = generate('H', 1.25, '1s,s', [0.9], 0.7, None, 2, 'PBE', True)
    assert G.check_all()
    basis = G.create_basis_set()
    basis.write_xml()
    setup_data = G.make_paw_setup('test')
    setup_data.write_xml()

    xc = XC('PBE')
    setup = Setup(setup_data, xc, lmax=2)
    T_Lqp = setup.calculate_T_Lqp(1, 3, 2,
                                  jlL_i=[(0, 0, 0), (1, 0, 0)])
    assert (T_Lqp[1:] == 0.0).all()
    assert T_Lqp[0] == pytest.approx(np.eye(3) / (4 * np.pi)**0.5)
    B_pqL = setup.xc_correction.B_pqL
    assert B_pqL == pytest.approx(T_Lqp.T)



from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
from ase import Atoms

from gpaw.calculator import GPAW
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.projections import Projections
from gpaw.setup import Setup
from gpaw.typing import Array2D, Array3D, Array4D, ArrayLike1D
from gpaw.utilities.ibz2bz import construct_symmetry_operators
from gpaw.utilities.partition import AtomPartition

from .functions import WannierFunctions


class WannierOverlaps:
    def __init__(self,
                 atoms: Atoms,
                 nwannier: int,
                 monkhorst_pack_size: ArrayLike1D,
                 kpoints: Array2D,
                 fermi_level: float,
                 directions: Dict[Tuple[int, ...], int],
                 overlaps: Array4D,
                 projections: Array3D = None,
                 proj_indices_a: List[List[int]] = None):

        self.atoms = atoms
        self.nwannier = nwannier
        self.monkhorst_pack_size = np.array(monkhorst_pack_size)
        self.kpoints = kpoints
        self.fermi_level = fermi_level
        self.directions = directions

        self.nkpts, ndirs, self.nbands, nbands = overlaps.shape
        assert nbands == self.nbands
        assert self.nkpts == np.prod(monkhorst_pack_size)  # type: ignore
        assert ndirs == len(directions)

        self._overlaps = overlaps
        self.projections = projections
        self.proj_indices_a = proj_indices_a

    def overlap(self,
                bz_index: int,
                direction: Tuple[int, ...]) -> Array2D:
        dindex = self.directions.get(direction)
        if dindex is not None:
            return self._overlaps[bz_index, dindex]

        size = self.monkhorst_pack_size
        i_c = np.unravel_index(bz_index, size)
        i2_c = np.array(i_c) + direction
        bz_index2 = np.ravel_multi_index(i2_c, size, 'wrap')  # type: ignore
        direction2 = tuple([-d for d in direction])
        dindex2 = self.directions[direction2]
        return self._overlaps[bz_index2, dindex2].T.conj()

    def localize_er(self,
                    maxiter: int = 100,
                    tolerance: float = 1e-5,
                    verbose: bool = not False) -> WannierFunctions:
        from .edmiston_ruedenberg import localize
        return localize(self, maxiter, tolerance, verbose)

    def localize_w90(self,
                     prefix: str = 'wannier',
                     folder: Union[Path, str] = 'W90',
                     nwannier: int = None,
                     **kwargs) -> WannierFunctions:
        from .w90 import Wannier90
        w90 = Wannier90(prefix, folder)
        w90.write_input_files(overlaps=self, **kwargs)
        w90.run_wannier90()
        return w90.read_result()


def dict_to_proj_indices(dct: Dict[Union[int, str], str],
                         setups: List[Setup]) -> List[List[int]]:
    """Convert dict to lists of projector function indices.

    >>> from gpaw.setup import create_setup
    >>> setup = create_setup('Si')  # 3s, 3p, *s, *p, *d
    >>> setup.n_j
    [3, 3, -1, -1, -1]
    >>> setup.l_j
    [0, 1, 0, 1, 2]
    >>> dict_to_proj_indices({'Si': 'sp', 1: 's'}, [setup, setup])
    [[0, 1, 2, 3], [0]]
    """
    indices_a = []
    for a, setup in enumerate(setups):
        ll = dct.get(a, dct.get(setup.symbol, ''))
        indices = []
        i = 0
        for n, l in zip(setup.n_j, setup.l_j):
            if n > 0 and 'spdf'[l] in ll:
                indices += list(range(i, i + 2 * l + 1))
            i += 2 * l + 1
        indices_a.append(indices)
    return indices_a


def calculate_overlaps(calc: GPAW,
                       nwannier: int,
                       projections: Dict[Union[int, str], str] = None,
                       n1: int = 0,
                       n2: int = 0,
                       spinors: bool = False,
                       spin: int = 0) -> WannierOverlaps:
    """Create WannierOverlaps object from DFT calculation.
    """
    assert not spinors

    if n2 <= 0:
        n2 += calc.get_number_of_bands()

    bzwfs = BZRealSpaceWaveFunctions.from_calculation(calc, n1, n2, spin)

    proj_indices_a = dict_to_proj_indices(projections or {},
                                          calc.setups)

    offsets = [0]
    for indices in proj_indices_a:
        offsets.append(offsets[-1] + len(indices))
    nproj = offsets.pop()

    if projections is not None:
        assert nproj == nwannier

    kd = bzwfs.kd
    gd = bzwfs.gd
    size = kd.N_c
    assert size is not None

    icell = calc.atoms.cell.reciprocal()
    directions = {direction: i
                  for i, direction
                  in enumerate(find_directions(icell, size))}
    Z_kdnn = np.empty((kd.nbzkpts, len(directions), n2 - n1, n2 - n1), complex)

    spos_ac = calc.spos_ac
    setups = calc.wfs.setups

    proj_kmn = np.zeros((kd.nbzkpts, nproj, n2 - n1), complex)

    for bz_index1 in range(kd.nbzkpts):
        wf1 = bzwfs[bz_index1]
        i1_c = np.unravel_index(bz_index1, size)
        for direction, d in directions.items():
            i2_c = np.array(i1_c) + direction
            bz_index2 = np.ravel_multi_index(i2_c,
                                             size,
                                             'wrap')  # type: ignore
            wf2 = bzwfs[bz_index2]
            phase_c = (i2_c % size - i2_c) // size
            u2_nR = wf2.u_nR
            if phase_c.any():
                u2_nR = u2_nR * gd.plane_wave(phase_c)
            Z_kdnn[bz_index1, d] = gd.integrate(wf1.u_nR, u2_nR,
                                                global_integral=False)

            for a, P1_ni in wf1.projections.items():
                dO_ii = setups[a].dO_ii
                P2_ni = wf2.projections[a]
                Z_nn = P1_ni.conj().dot(dO_ii).dot(P2_ni.T).astype(complex)
                if phase_c.any():
                    Z_nn *= np.exp(2j * np.pi * phase_c.dot(spos_ac[a]))
                Z_kdnn[bz_index1, d] += Z_nn

        for a, P1_ni in wf1.projections.items():
            indices = proj_indices_a[a]
            m = offsets[a]
            proj_kmn[bz_index1, m:m + len(indices)] = P1_ni.T[indices]

    gd.comm.sum(Z_kdnn)
    gd.comm.sum(proj_kmn)

    overlaps = WannierOverlaps(calc.atoms,
                               nwannier,
                               size,
                               kd.bzk_kc,
                               calc.get_fermi_level(),
                               directions,
                               Z_kdnn,
                               proj_kmn,
                               proj_indices_a)
    return overlaps


def find_directions(icell: Array2D,
                    mpsize: ArrayLike1D) -> List[Tuple[int, ...]]:
    """Find nearest neighbors k-points.

    icell:
        Reciprocal cell.
    mpsize:
        Size of Monkhorst-Pack grid.

    If dk is a vector pointing at a neighbor k-points then we don't
    also include -dk in the list.  Examples: for simple cubic there
    will be 3 neighbors and for FCC there will be 6.

    For a hexagonal cell you get three directions in plane and one
    out of plane:

    >>> hex = np.array([[1, 0, 0], [0.5, 3**0.5 / 2, 0], [0, 0, 1]])
    >>> dirs = find_directions(hex, (4, 4, 4))
    >>> sorted(dirs)
    [(0, 0, 1), (0, 1, 0), (1, -1, 0), (1, 0, 0)]
    """

    from scipy.spatial import Voronoi

    d_ic = np.indices((3, 3, 3)).reshape((3, -1)).T - 1
    d_iv = d_ic.dot((icell.T / mpsize).T)
    voro = Voronoi(d_iv)
    directions: List[Tuple[int, ...]] = []
    for i1, i2 in voro.ridge_points:
        if i1 == 13 and i2 > 13:
            directions.append(tuple(d_ic[i2]))
        elif i2 == 13 and i1 > 13:
            directions.append(tuple(d_ic[i1]))
    return directions


class WaveFunction:
    def __init__(self,
                 u_nR,
                 projections: Projections):
        self.u_nR = u_nR
        self.projections = projections

    def transform(self,
                  kd: KPointDescriptor,
                  gd,
                  setups: List[Setup],
                  spos_ac: Array2D,
                  bz_index: int) -> 'WaveFunction':
        """Transforms PAW projections from IBZ to BZ k-point."""
        a_a, U_aii, time_rev = construct_symmetry_operators(
            kd, setups, spos_ac, bz_index)

        projections = self.projections.new()

        if projections.atom_partition.comm.rank == 0:
            a = 0
            for b, U_ii in zip(a_a, U_aii):
                P_msi = self.projections[b].dot(U_ii)
                if time_rev:
                    P_msi = P_msi.conj()
                projections[a][:] = P_msi
                a += 1
        else:
            assert len(projections.indices) == 0

        u_nR = 1 / 0
        return WaveFunction(u_nR, projections)

    def redistribute_atoms(self,
                           gd,
                           atom_partition: AtomPartition
                           ) -> 'WaveFunction':
        projections = self.projections.redist(atom_partition)
        u_nR = gd.distribute(self.u_nR)
        return WaveFunction(u_nR, projections)


class BZRealSpaceWaveFunctions:
    """Container for wave-functions and PAW projections (all of BZ)."""
    def __init__(self,
                 kd: KPointDescriptor,
                 gd,
                 wfs: Dict[int, WaveFunction]):
        self.kd = kd
        self.gd = gd
        self.wfs = wfs

    def __getitem__(self, bz_index):
        return self.wfs[bz_index]

    @classmethod
    def from_calculation(cls,
                         calc: GPAW,
                         n1: int = 0,
                         n2: int = 0,
                         spin=0) -> 'BZRealSpaceWaveFunctions':
        wfs = calc.wfs
        kd = wfs.kd

        if wfs.mode == 'lcao' and not wfs.positions_set:
            calc.initialize_positions()

        gd = wfs.gd.new_descriptor(comm=calc.world)

        nproj_a = wfs.kpt_qs[0][0].projections.nproj_a
        # All atoms on rank-0:
        rank_a = np.zeros_like(nproj_a)
        atom_partition = AtomPartition(gd.comm, rank_a)

        rank_a = np.arange(len(rank_a)) % gd.comm.size
        atom_partition2 = AtomPartition(gd.comm, rank_a)

        spos_ac = calc.spos_ac
        setups = calc.wfs.setups

        u_nR = gd.empty((n2 - n1), complex, global_array=True)

        bzwfs = {}
        for ibz_index in range(kd.nibzkpts):
            for n in range(n1, n2):
                u_nR[n - n1] = wfs.get_wave_function_array(n=n,
                                                           k=ibz_index,
                                                           s=spin,
                                                           periodic=True)
            P_nI = wfs.collect_projections(ibz_index, spin)
            if P_nI is not None:
                P_nI = P_nI[n1:n2]
            projections = Projections(
                nbands=n2 - n1,
                nproj_a=nproj_a,
                atom_partition=atom_partition,
                data=P_nI)

            wf = WaveFunction(u_nR.copy(), projections)

            for bz_index, ibz_index2 in enumerate(kd.bz2ibz_k):
                if ibz_index2 != ibz_index:
                    continue
                if kd.ibz2bz_k[ibz_index] == bz_index:
                    wf1 = wf
                else:
                    wf1 = wf.transform(kd, gd, setups, spos_ac, bz_index)

                bzwfs[bz_index] = wf1.redistribute_atoms(gd, atom_partition2)

        return BZRealSpaceWaveFunctions(kd, gd, bzwfs)



from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import world


def test_pw_direct():
    if world.size == 1:
        atoms = Atoms('H', cell=(2, 2, 2), pbc=True)
        atoms.calc = GPAW(mode=PW(300, force_complex_dtype=True),
                          eigensolver='direct')
        atoms.get_potential_energy()

    if world.size == 2:
        atoms = Atoms('H', cell=(2, 2, 2), pbc=True)
        atoms.calc = GPAW(mode=PW(300),
                          eigensolver='direct',
                          kpts=(3, 2, 2))
        atoms.get_potential_energy()



from gpaw.utilities.blas import rk
from gpaw.utilities.blas import mmm


def syrk(trans, a, c, alpha, beta, lower):
    assert trans == 'N'
    assert lower
    rk(alpha, a._data, beta, c._data)


def gemm(transa, transb, a, b, c, alpha, beta):
    mmm(alpha,
        a._data,
        transa.replace('H', 'C'),
        b._data,
        transb.replace('H', 'C'),
        beta,
        c._data)



import argparse
import runpy
import sys


class CLICommand:
    """Run GPAW's Python interpreter."""

    @staticmethod
    def add_arguments(parser):
        parser.add_argument(
            '--dry-run', '-z', type=int, default=0,
            metavar='NCPUS',
            help='Dry run on NCPUS cpus.')
        parser.add_argument(
            '-d', '--debug', action='store_true',
            help='Run in debug-mode.')

        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--command', '-c',
            dest='cmd',
            nargs=argparse.REMAINDER,
            help='Program passed in as string (terminates option list).')
        group.add_argument(
            '--module', '-m',
            nargs=argparse.REMAINDER,
            help='Run library module as a script (terminates option list).')

        parser.add_argument(
            'arguments', metavar='ARG',
            help='Arguments passed to program in '
            'sys.argv[1:].  '
            'Use -- to force all remaining arguments to be '
            'passed to target program (useful for passing '
            'options starting with -).',
            nargs='*')

    @staticmethod
    def run(args):
        if args.cmd:
            sys.argv[:] = ['-c'] + args.cmd[1:]
            d = {}
            exec(args.cmd[0], d, d)
        elif args.module:
            sys.argv[:] = args.module
            runpy.run_module(args.module[0],
                             run_name='__main__',
                             alter_sys=True)
        else:
            sys.argv[:] = args.arguments
            runpy.run_path(args.arguments[0], run_name='__main__')



from __future__ import annotations
from gpaw.occupations import create_occ_calc, ParallelLayout
from gpaw.band_descriptor import BandDescriptor
from gpaw.typing import ArrayLike2D, Array2D


class OccupationNumberCalculator:
    def __init__(self,
                 dct,
                 pbc,
                 ibz,
                 nbands,
                 comms,
                 magmom_v,
                 ncomponents,
                 rcell):
        if dct is None:
            if pbc.any():
                dct = {'name': 'fermi-dirac',
                       'width': 0.1}  # eV
            else:
                dct = {'width': 0.0}

        if dct.get('fixmagmom'):
            if ncomponents == 1:
                dct = dct.copy()
                del dct['fixmagmom']
            assert ncomponents == 2

        kwargs = dct.copy()
        name = kwargs.pop('name', '')
        if name == 'mom':
            1 / 0
            # from gpaw.mom import OccupationsMOM
            # return OccupationsMOM(..., **kwargs)

        self.band_comm = comms['b']
        bd = BandDescriptor(nbands)
        self.occ = create_occ_calc(
            dct,
            parallel_layout=ParallelLayout(bd,
                                           comms['k'],
                                           comms['d']),
            fixed_magmom_value=magmom_v[2],
            rcell=rcell,
            monkhorst_pack_size=getattr(ibz.bz, 'size_c', None),
            bz2ibzmap=ibz.bz2ibz_K)
        self.extrapolate_factor = self.occ.extrapolate_factor

    def __str__(self):
        return str(self.occ)

    def calculate(self,
                  nelectrons: float,
                  eigenvalues: ArrayLike2D,
                  weights: list[float],
                  fermi_levels_guess: list[float] = None
                  ) -> tuple[Array2D, list[float], float]:
        if self.band_comm.rank == 0:
            occs, fls, e = self.occ.calculate(nelectrons, eigenvalues, weights,
                                              fermi_levels_guess)
        else:
            1 / 0
        return occs, fls, e



from math import log

import pytest
from ase import Atoms
from ase.db import connect
from ase.io import iread, read, write
from ase.units import Bohr

from gpaw import GPAW, FermiDirac
from gpaw.test import equal


@pytest.mark.later
def test_generic_hydrogen(in_tmp_dir):
    a = 4.0
    h = 0.2
    hydrogen = Atoms('H',
                     [(a / 2, a / 2, a / 2)],
                     cell=(a, a, a))

    params = dict(h=h,
                  nbands=1,
                  convergence={'energy': 1e-7})
    hydrogen.calc = GPAW(**params,
                         txt='h.txt')
    e1 = hydrogen.get_potential_energy()
    equal(e1, 0.526939, 0.001)

    dens = hydrogen.calc.density
    c = dens.gd.find_center(dens.nt_sG[0]) * Bohr
    equal(abs(c - a / 2).max(), 0, 1e-13)

    kT = 0.001
    hydrogen.calc = GPAW(**params,
                         occupations=FermiDirac(width=kT))
    e2 = hydrogen.get_potential_energy()
    equal(e1, e2 + log(2) * kT, 3.0e-7)

    # Test ase.db a bit:
    for name in ['h.json', 'h.db']:
        con = connect(name, append=False)
        con.write(hydrogen)
        id = con.write(hydrogen, foo='bar', data={'abc': [1, 2, 3]})
        assert id == 2
        assert con.reserve(foo='bar') is None
        row = con.get(foo='bar')
        assert row.energy == e2
        assert sum(row.data.abc) == 6
        del con[1]
        assert con.reserve(x=42) == 3

        write('x' + name, hydrogen)
        write('xx' + name, [hydrogen, hydrogen])

        assert read(name + '@foo=bar')[0].get_potential_energy() == e2
        for n, h in zip([1, 0], iread(name + '@:')):
            assert n == len(h)

    # Test parsing of GPAW's text output:
    h = read('h.txt')
    error = abs(h.calc.get_eigenvalues() -
                hydrogen.calc.get_eigenvalues()).max()
    assert error < 1e-5, error

    # Test get_electrostatic_potential() method
    v = hydrogen.calc.get_electrostatic_potential()
    print(v.shape, v.ptp())



import numpy as np
from math import pi
from gpaw.response.q0_correction import Q0Correction
from ase.units import Ha
from ase.dft.kpoints import monkhorst_pack
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.response.temp import DielectricFunctionCalculator
from gpaw.response.hilbert import GWHilbertTransforms


class QPointDescriptor(KPointDescriptor):

    @staticmethod
    def from_gs(gs):
        kd, atoms = gs.kd, gs.atoms
        # Find q-vectors and weights in the IBZ:
        assert -1 not in kd.bz2bz_ks
        offset_c = 0.5 * ((kd.N_c + 1) % 2) / kd.N_c
        bzq_qc = monkhorst_pack(kd.N_c) + offset_c
        qd = KPointDescriptor(bzq_qc)
        qd.set_symmetry(atoms, kd.symmetry)
        return qd


def initialize_w_calculator(chi0calc, context, *,
                            coulomb,
                            xc='RPA',  # G0W0Kernel arguments
                            ppa=False, E0=Ha, eta=None,
                            integrate_gamma=0, q0_correction=False):
    """Initialize a WCalculator from a Chi0Calculator.

    Parameters
    ----------
    chi0calc : Chi0Calculator
    xc : str
        Kernel to use when including vertex corrections.

    Remaining arguments: See WCalculator
    """
    from gpaw.response.g0w0_kernels import G0W0Kernel

    gs = chi0calc.gs
    qd = QPointDescriptor.from_gs(gs)

    xckernel = G0W0Kernel(xc=xc, ecut=chi0calc.ecut,
                          gs=gs, qd=qd,
                          ns=gs.nspins,
                          context=context)
    if ppa:
        wcalc = PPACalculator(gs, context, qd=qd,
                              coulomb=coulomb, xckernel=xckernel,
                              integrate_gamma=integrate_gamma, eta=eta,
                              q0_correction=q0_correction)
    else:
        if eta is not None:
            hilbert_transform = GWHilbertTransforms(chi0calc.wd.omega_w, eta)
        else:
            hilbert_transform = None
        wcalc = WCalculator(gs, context, qd=qd,
                            coulomb=coulomb, xckernel=xckernel,
                            integrate_gamma=integrate_gamma,
                            hilbert_transform=hilbert_transform,
                            q0_correction=q0_correction)

    return wcalc


class WBaseCalculator():

    def __init__(self, gs, context, *, qd,
                 coulomb, xckernel, integrate_gamma=0, q0_correction=False):
        """
        Base class for W Calculator including basic initializations and Gamma
        Gamma handling.

        Parameters
        ----------
        gs : ResponseGroundStateAdapter
        context : ResponseContext
        qd : QPointDescriptor
        coulomb : CoulombKernel
        xckernel : G0W0Kernel
        integrate_gamma: int
             Method to integrate the Coulomb interaction. 1 is a numerical
             integration at all q-points with G=[0,0,0] - this breaks the
             symmetry slightly. 0 is analytical integration at q=[0,0,0] only
             this conserves the symmetry. integrate_gamma=2 is the same as 1,
             but the average is only carried out in the non-periodic directions
        q0_correction : bool
            Analytic correction to the q=0 contribution applicable to 2D
            systems.
        """
        self.gs = gs
        self.context = context
        self.qd = qd
        self.coulomb = coulomb
        self.xckernel = xckernel
        self.integrate_gamma = integrate_gamma

        if q0_correction:
            assert self.coulomb.truncation == '2D'
            self.q0_corrector = Q0Correction(
                cell_cv=self.gs.gd.cell_cv,
                bzk_kc=self.gs.kd.bzk_kc,
                N_c=self.qd.N_c)

            npts_c = self.q0_corrector.npts_c
            self.context.print('Applying analytical 2D correction to W:',
                               flush=False)
            self.context.print('    Evaluating Gamma point contribution to W '
                               + 'on a %dx%dx%d grid' % tuple(npts_c))
        else:
            self.q0_corrector = None
    
    def get_V0sqrtV0(self, chi0):
        """
        Integrated Coulomb kernels.
        integrate_gamma = 0: Analytically integrated kernel
        in sphere around Gamma
        integrate_gamma > 0: Numerically  integrated kernel
        XXX: Understand and document Rq0, V0, sqrtV0
        """
        V0 = None
        sqrtV0 = None
        if self.integrate_gamma != 0:
            reduced = (self.integrate_gamma == 2)
            V0, sqrtV0 = self.coulomb.integrated_kernel(qpd=chi0.qpd,
                                                        reduced=reduced)
        elif self.integrate_gamma == 0 and chi0.optical_limit:
            bzvol = (2 * np.pi)**3 / self.gs.volume / self.qd.nbzkpts
            Rq0 = (3 * bzvol / (4 * np.pi))**(1. / 3.)
            V0 = 16 * np.pi**2 * Rq0 / bzvol
            sqrtV0 = (4 * np.pi)**(1.5) * Rq0**2 / bzvol / 2
        return V0, sqrtV0

    def apply_gamma_correction(self, W_GG, einv_GG, V0, sqrtV0, sqrtV_G):
        """
        Replacing q=0, (G,G')= (0,0), (0,:), (:,0) with corresponding
        matrix elements calculated with an average of the (diverging)
        Coulomb interaction.
        XXX: Understand and document exact expressions
        """
        W_GG[0, 0] = einv_GG[0, 0] * V0
        W_GG[0, 1:] = einv_GG[0, 1:] * sqrtV_G[1:] * sqrtV0
        W_GG[1:, 0] = einv_GG[1:, 0] * sqrtV0 * sqrtV_G[1:]

    
class WCalculator(WBaseCalculator):
    def __init__(self, gs, context, *, qd,
                 coulomb, xckernel, hilbert_transform,
                 integrate_gamma, q0_correction):
        super().__init__(gs, context, qd=qd, coulomb=coulomb,
                         xckernel=xckernel,
                         integrate_gamma=integrate_gamma,
                         q0_correction=q0_correction)
        self.hilbert_transform = hilbert_transform

    def get_HW_model(self, chi0, fxc_mode, only_correlation=True):
        assert only_correlation
        W_wGG = self.calculate_W_WgG(chi0,
                                     fxc_mode=fxc_mode,
                                     only_correlation=True)
        # HT used to calculate convulution between time-ordered G and W
        with self.context.timer('Hilbert'):
            W_xwGG = self.hilbert_transform(W_wGG)

        factor = 1.0 / (self.qd.nbzkpts * 2 * pi * self.gs.volume)
        return FullFrequencyHWModel(chi0.wd, W_xwGG, factor)
        
    def calculate_W_WgG(self, chi0,
                        fxc_mode='GW',
                        only_correlation=False):
        """Calculate the screened interaction in W_wGG or W_WgG representation.

        Additional Parameters
        ----------
        only_correlation: bool
             if true calculate Wc otherwise calculate full W
        out_dist: str
             specifices output distribution of W array (wGG or WgG)
        """
        W_wGG = self.calculate_W_wGG(chi0, fxc_mode,
                                     only_correlation=only_correlation)
        
        W_WgG = chi0.blockdist.distribute_as(W_wGG, chi0.nw, 'WgG')
        return W_WgG

    def calculate_W_wGG(self, chi0, fxc_mode='GW',
                        only_correlation=False):
        """In-place calculation of the screened interaction."""
        chi0_wGG = chi0.copy_array_with_distribution('wGG')
        dfc = DielectricFunctionCalculator(chi0, self.coulomb,
                                           self.xckernel, fxc_mode)
        self.context.timer.start('Dyson eq.')
        
        V0, sqrtV0 = self.get_V0sqrtV0(chi0)
        for iw, chi0_GG in enumerate(chi0_wGG):
            # Note, at q=0 get_epsinv_GG modifies chi0_GG
            einv_GG = dfc.get_epsinv_GG(chi0_GG, iw)
            # Renaming the chi0_GG buffer since it will be used to store W
            W_GG = chi0_GG
            # If only_correlation = True function spits out
            # W^c = sqrt(V)(epsinv - delta_GG')sqrt(V). However, full epsinv
            # is still needed for q0_corrector.
            einvt_GG = (einv_GG - dfc.I_GG) if only_correlation else einv_GG
            W_GG[:] = einvt_GG * (dfc.sqrtV_G *
                                  dfc.sqrtV_G[:, np.newaxis])
            if self.q0_corrector is not None and chi0.optical_limit:
                W = dfc.wblocks1d.a + iw
                self.q0_corrector.add_q0_correction(chi0.qpd, W_GG,
                                                    einv_GG,
                                                    chi0.chi0_WxvG[W],
                                                    chi0.chi0_Wvv[W],
                                                    dfc.sqrtV_G)
                # XXX Is it to correct to have "or" here?
            elif chi0.optical_limit or self.integrate_gamma != 0:
                self.apply_gamma_correction(W_GG, einvt_GG,
                                            V0, sqrtV0, dfc.sqrtV_G)

        self.context.timer.stop('Dyson eq.')
        return chi0_wGG

    def dyson_and_W_new(self, iq, q_c, chi0, ecut, coulomb):
        # assert not self.ppa
        # assert not self.do_GW_too
        assert ecut == chi0.qpd.ecut
        assert self.fxc_mode == 'GW'
        assert not np.allclose(q_c, 0)

        nW = len(self.wd)
        nG = chi0.qpd.ngmax

        from gpaw.response.wgg import Grid

        WGG = (nW, nG, nG)
        WgG_grid = Grid(
            comm=self.blockcomm,
            shape=WGG,
            cpugrid=(1, self.blockcomm.size, 1))
        assert chi0.chi0_wGG.shape == WgG_grid.myshape

        my_gslice = WgG_grid.myslice[1]

        dielectric_WgG = chi0.chi0_wGG  # XXX
        for iw, chi0_GG in enumerate(chi0.chi0_wGG):
            sqrtV_G = coulomb.sqrtV(chi0.qpd, q_v=None)
            e_GG = np.eye(nG) - chi0_GG * sqrtV_G * sqrtV_G[:, np.newaxis]
            e_gG = e_GG[my_gslice]

            dielectric_WgG[iw, :, :] = e_gG

        wgg_grid = Grid(comm=self.blockcomm, shape=WGG)

        dielectric_wgg = wgg_grid.zeros(dtype=complex)
        WgG_grid.redistribute(wgg_grid, dielectric_WgG, dielectric_wgg)

        assert np.allclose(dielectric_wgg, dielectric_WgG)

        wgg_grid.invert_inplace(dielectric_wgg)

        wgg_grid.redistribute(WgG_grid, dielectric_wgg, dielectric_WgG)
        inveps_WgG = dielectric_WgG

        self.context.timer.start('Dyson eq.')

        for iw, inveps_gG in enumerate(inveps_WgG):
            inveps_gG -= np.identity(nG)[my_gslice]
            thing_GG = sqrtV_G * sqrtV_G[:, np.newaxis]
            inveps_gG *= thing_GG[my_gslice]

        W_WgG = inveps_WgG
        Wp_wGG = W_WgG.copy()
        Wm_wGG = W_WgG.copy()
        return chi0.qpd, Wm_wGG, Wp_wGG  # not Hilbert transformed yet


class HWModel:
    """
        Hilbert Transformed W Model.
    """

    def get_HW(self, omega, fsign):
        """
            Get Hilbert transformed W at frequency omega.

            The fsign is utilize to select which type of Hilbert transform
            is selected, as is detailed in Sigma expectation value evaluation
            where this model is used.
        """
        raise NotImplementedError


class FullFrequencyHWModel(HWModel):
    def __init__(self, wd, HW_swGG, factor):
        self.wd = wd
        self.HW_swGG = HW_swGG
        self.factor = factor

    def get_HW(self, omega, fsign):
        # For more information about how fsign, and wsign works, see
        # https://backend.orbit.dtu.dk/ws/portalfiles/portal/93075765/hueser_PhDthesis.pdf
        # eq. 2.2 endind up to eq. 2.11
        # Effectively, the symmetry of time ordered W is used,
        # i.e. W(w) = -W(-w). To allow that data is only stored for w>=0.
        # Hence, the interpolation happends always to the positive side, but
        # the information of true w is keps tract using wsign.
        # In addition, whether the orbital in question at G is occupied or
        # unoccupied, which then again affects, which Hilbert transform of
        # W is chosen, is kept track with fsign.
        o = abs(omega)
        wsign = np.sign(omega + 1e-15)
        wd = self.wd
        # Pick +i*eta or -i*eta:
        s = (1 + wsign * np.sign(-fsign)).astype(int) // 2
        w = wd.get_floor_index(o, safe=False)
      
        # Interpolation indexes w + 1, therefore - 2 here
        if w > len(wd) - 2:
            return None, None

        o1 = wd.omega_w[w]
        o2 = wd.omega_w[w + 1]

        C1_GG = self.HW_swGG[s][w]
        C2_GG = self.HW_swGG[s][w + 1]
        p = self.factor * wsign

        sigma_GG = ((o - o1) * C2_GG + (o2 - o) * C1_GG) / (o2 - o1)
        dsigma_GG = wsign * (C2_GG - C1_GG) / (o2 - o1)
        return -1j * p * sigma_GG, -1j * p * dsigma_GG


class PPAHWModel(HWModel):
    def __init__(self, W_GG, omegat_GG, eta, factor):
        self.W_GG = W_GG
        self.omegat_GG = omegat_GG
        self.eta = eta
        self.factor = factor

    def get_HW(self, omega, sign):
        omegat_GG = self.omegat_GG
        W_GG = self.W_GG

        x1_GG = 1 / (omega + omegat_GG - 1j * self.eta)
        x2_GG = 1 / (omega - omegat_GG + 1j * self.eta)
        x3_GG = 1 / (omega + omegat_GG - 1j * self.eta * sign)
        x4_GG = 1 / (omega - omegat_GG - 1j * self.eta * sign)
        x_GG = self.factor * W_GG * (sign * (x1_GG - x2_GG) + x3_GG + x4_GG)
        dx_GG = -self.factor * W_GG * (sign * (x1_GG**2 - x2_GG**2) +
                                       x3_GG**2 + x4_GG**2)
        return x_GG, dx_GG


class PPACalculator(WBaseCalculator):
    def __init__(self, gs, context, *, qd,
                 coulomb, xckernel,
                 integrate_gamma, q0_correction, eta):
        super().__init__(gs, context, qd=qd, coulomb=coulomb,
                         xckernel=xckernel,
                         integrate_gamma=integrate_gamma,
                         q0_correction=q0_correction)
        self.eta = eta

    def get_HW_model(self, chi0,
                     fxc_mode='GW'):
        """Calculate the PPA parametrization of screened interaction.
        """
        assert len(chi0.wd.omega_w) == 2
        # E0 directly related to frequency mesh for chi0
        E0 = chi0.wd.omega_w[1].imag

        dfc = DielectricFunctionCalculator(chi0,
                                           self.coulomb,
                                           self.xckernel,
                                           fxc_mode)

        V0, sqrtV0 = self.get_V0sqrtV0(chi0)
        self.context.timer.start('Dyson eq.')
        einv_wGG = dfc.get_epsinv_wGG(only_correlation=True)
        omegat_GG = E0 * np.sqrt(einv_wGG[1] /
                                 (einv_wGG[0] - einv_wGG[1]))
        R_GG = -0.5 * omegat_GG * einv_wGG[0]
        W_GG = pi * R_GG * dfc.sqrtV_G * dfc.sqrtV_G[:, np.newaxis]
        if chi0.optical_limit or self.integrate_gamma != 0:
            self.apply_gamma_correction(W_GG, pi * R_GG,
                                        V0, sqrtV0,
                                        dfc.sqrtV_G)

        self.context.timer.stop('Dyson eq.')

        factor = 1.0 / (self.qd.nbzkpts * 2 * pi * self.gs.volume)

        return PPAHWModel(W_GG, omegat_GG, self.eta, factor)



import numpy as np


def generate_gridspec(**kwargs):
    from matplotlib.gridspec import GridSpec
    width = 0.84
    bottom = 0.12
    left = 0.12
    return GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[1, 3],
                    bottom=bottom, top=bottom + width,
                    left=left, right=left + width,
                    **kwargs)


def plot_DOS(ax, energy_e, dos_e, base_e, dos_min, dos_max,
             flip=False, fill=None, line=None):
    ax.xaxis.set_ticklabels([])
    ax.yaxis.set_ticklabels([])
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.yaxis.set_ticks_position('left')
    ax.xaxis.set_ticks_position('bottom')
    if flip:
        set_label = ax.set_xlabel
        fill_between = ax.fill_betweenx
        set_energy_lim = ax.set_ylim
        set_dos_lim = ax.set_xlim

        def plot(x, y, *args, **kwargs):
            return ax.plot(y, x, *args, **kwargs)
    else:
        set_label = ax.set_ylabel
        fill_between = ax.fill_between
        set_energy_lim = ax.set_xlim
        set_dos_lim = ax.set_ylim

        def plot(x, y, *args, **kwargs):
            return ax.plot(x, y, *args, **kwargs)
    if fill:
        fill_between(energy_e, base_e, dos_e + base_e, **fill)
    if line:
        plot(energy_e, dos_e, **line)
    set_label('DOS', labelpad=0)
    set_energy_lim(np.take(energy_e, (0, -1)))
    set_dos_lim(dos_min, dos_max)


class TCM(object):

    def __init__(self, energy_o, energy_u, fermilevel):
        self.energy_o = energy_o
        self.energy_u = energy_u
        self.fermilevel = fermilevel

        self.base_o = np.zeros_like(energy_o)
        self.base_u = np.zeros_like(energy_u)

    def __getattr__(self, attr):
        import matplotlib.pyplot as plt
        # Generate axis only when needed
        if attr in ['ax_occ_dos', 'ax_unocc_dos', 'ax_tcm']:
            gs = generate_gridspec(hspace=0.05, wspace=0.05)
            self.ax_occ_dos = plt.subplot(gs[0])
            self.ax_unocc_dos = plt.subplot(gs[3])
            self.ax_tcm = plt.subplot(gs[2])
            return getattr(self, attr)
        if attr in ['ax_spec']:
            gs = generate_gridspec(hspace=0.8, wspace=0.8)
            self.ax_spec = plt.subplot(gs[1])
            return getattr(self, attr)
        if attr in ['ax_cbar']:
            self.ax_cbar = plt.axes((0.15, 0.6, 0.02, 0.1))
            return getattr(self, attr)
        raise AttributeError('%s object has no attribute %s' %
                             (repr(self.__class__.__name__), repr(attr)))

    def plot_TCM(self, tcm_ou, vmax='80%', vmin='symmetrize', cmap='seismic',
                 log=False, colorbar=True, lw=None):
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from matplotlib.colors import Normalize, LogNorm
        if lw is None:
            lw = mpl.rcParams['lines.linewidth']
        energy_o = self.energy_o
        energy_u = self.energy_u
        fermilevel = self.fermilevel

        tcmmax = np.max(np.absolute(tcm_ou))
        print('tcmmax', tcmmax)

        # Plot TCM
        ax = self.ax_tcm
        plt.sca(ax)
        plt.cla()
        if isinstance(vmax, str):
            assert vmax[-1] == '%'
            tcmmax = np.max(np.absolute(tcm_ou))
            vmax = tcmmax * float(vmax[:-1]) / 100.0
        if vmin == 'symmetrize':
            vmin = -vmax
        if tcm_ou.dtype == complex:
            linecolor = 'w'
            from matplotlib.colors import hsv_to_rgb

            def transform_to_hsv(z, rmin, rmax, hue_start=90):
                amp = np.absolute(z)  # **2
                amp = np.where(amp < rmin, rmin, amp)
                amp = np.where(amp > rmax, rmax, amp)
                ph = np.angle(z, deg=1) + hue_start
                h = (ph % 360) / 360
                s = 1.85 * np.ones_like(h)
                v = (amp - rmin) / (rmax - rmin)
                return hsv_to_rgb(np.dstack((h, s, v)))

            img = transform_to_hsv(tcm_ou.T, 0, vmax)
            plt.imshow(img, origin='lower',
                       extent=[energy_o[0], energy_o[-1],
                               energy_u[0], energy_u[-1]],
                       interpolation='bilinear',
                       )
        else:
            linecolor = 'k'
            if cmap == 'magma':
                linecolor = 'w'
            if log:
                norm = LogNorm(vmin=vmin, vmax=vmax)
            else:
                norm = Normalize(vmin=vmin, vmax=vmax)
            plt.pcolormesh(energy_o, energy_u, tcm_ou.T,
                           cmap=cmap, rasterized=True, norm=norm,
                           shading='auto')
        if colorbar:
            ax = self.ax_cbar
            ax.clear()
            cb = plt.colorbar(cax=ax)
            cb.outline.set_edgecolor(linecolor)
            ax.tick_params(axis='both', colors=linecolor)
            # ax.yaxis.label.set_color(linecolor)
            # ax.xaxis.label.set_color(linecolor)
        ax = self.ax_tcm
        plt.sca(ax)
        plt.axhline(fermilevel, c=linecolor, lw=lw)
        plt.axvline(fermilevel, c=linecolor, lw=lw)

        ax.tick_params(axis='both', which='major', pad=2)
        plt.xlabel(r'Occ. energy $\varepsilon_{o}$ (eV)', labelpad=0)
        plt.ylabel(r'Unocc. energy $\varepsilon_{u}$ (eV)', labelpad=0)
        plt.xlim(np.take(energy_o, (0, -1)))
        plt.ylim(np.take(energy_u, (0, -1)))

    def plot_DOS(self, dos_o, dos_u, stack=False,
                 fill={'color': '0.8'}, line={'color': 'k'}):
        # Plot DOSes
        if stack:
            base_o = self.base_o
            base_u = self.base_u
        else:
            base_o = np.zeros_like(self.energy_o)
            base_u = np.zeros_like(self.energy_u)
        dos_min = 0.0
        dos_max = 1.01 * max(np.max(dos_o), np.max(dos_u))
        plot_DOS(self.ax_occ_dos, self.energy_o, dos_o, base_o,
                 dos_min, dos_max, flip=False, fill=fill, line=line)
        plot_DOS(self.ax_unocc_dos, self.energy_u, dos_u, base_u,
                 dos_min, dos_max, flip=True, fill=fill, line=line)
        if stack:
            self.base_o += dos_o
            self.base_u += dos_u

    def plot_spectrum(self):
        raise NotImplementedError()

    def plot_TCM_diagonal(self, energy, **kwargs):
        x_o = np.take(self.energy_o, (0, -1))
        self.ax_tcm.plot(x_o, x_o + energy, **kwargs)

    def set_title(self, *args, **kwargs):
        self.ax_occ_dos.set_title(*args, **kwargs)


class TCMPlotter(TCM):

    def __init__(self, ksd, energy_o, energy_u, sigma,
                 zero_fermilevel=True):
        eig_n, fermilevel = ksd.get_eig_n(zero_fermilevel)
        TCM.__init__(self, energy_o, energy_u, fermilevel)
        self.ksd = ksd
        self.sigma = sigma
        self.eig_n = eig_n

    def plot_TCM(self, weight_p, **kwargs):
        # Calculate TCM
        tcm_ou = self.ksd.get_TCM(weight_p, self.eig_n, self.energy_o,
                                  self.energy_u, self.sigma)
        TCM.plot_TCM(self, tcm_ou, **kwargs)

    def plot_DOS(self, weight_n=1.0, **kwargs):
        # Calculate DOS
        dos_o, dos_u = self.ksd.get_weighted_DOS(weight_n, self.eig_n,
                                                 self.energy_o,
                                                 self.energy_u,
                                                 self.sigma)
        TCM.plot_DOS(self, dos_o, dos_u, **kwargs)



from abc import ABC, abstractmethod
from time import time
from pathlib import Path

import ase.io.ulm as ulm
import numpy as np
from ase.units import Ha
from gpaw.response import timer
from scipy.special import p_roots, sici

from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.fd_operators import Gradient
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.utilities.blas import axpy, gemmdot
from gpaw.xc.rpa import RPACorrelation
from gpaw.heg import HEG
from gpaw.xc.fxc_kernels import (
    get_fHxc_Gr, get_pbe_fxc, get_fspinHxc_Gr_rALDA, get_fspinHxc_Gr_rAPBE)


def get_chi0v_spinsum(chi0_sGG, G_G):
    nG = chi0_sGG.shape[-1]
    chi0v = np.zeros((nG, nG), dtype=complex)
    for chi0_GG in chi0_sGG:
        chi0v += chi0_GG / G_G / G_G[:, np.newaxis]
    chi0v *= 4 * np.pi
    return chi0v


def get_chi0v_foreach_spin(chi0_sGG, G_G):
    ns, nG = chi0_sGG.shape[:2]

    chi0v_sGsG = np.zeros((ns, nG, ns, nG), dtype=complex)
    for s in range(ns):
        chi0v_sGsG[s, :, s, :] = chi0_sGG[s] / G_G / G_G[:, np.newaxis]
    chi0v_sGsG *= 4 * np.pi
    return chi0v_sGsG.reshape(ns * nG, ns * nG)


class FXCCache:
    def __init__(self, comm, tag, xc, ecut):
        self.comm = comm
        self.tag = tag
        self.xc = xc
        self.ecut = ecut

    @property
    def prefix(self):
        return f'{self.tag}_{self.xc}_{self.ecut}'

    def handle(self, iq):
        return Handle(self, iq)


class Handle:
    def __init__(self, cache, iq):
        self.cache = cache
        self.iq = iq
        self.comm = cache.comm

    @property
    def _path(self):
        return Path(f'fhxc_{self.cache.prefix}_{self.iq}.ulm')

    def exists(self):
        if self.comm.rank == 0:
            exists = int(self._path.exists())
        else:
            exists = 0
        exists = self.comm.sum_scalar(exists)
        return bool(exists)

    def read(self):
        from gpaw.mpi import broadcast
        if self.comm.rank == 0:
            array = self._read_master()
            assert array is not None
        else:
            array = None

        # The shape of the array is only known on rank0,
        # so we cannot use the in-place broadcast.  Therefore
        # we use the standalone function.
        array = broadcast(array, root=0, comm=self.comm)
        assert array is not None
        return array

    def write(self, array):
        if self.comm.rank == 0:
            assert array is not None
            self._write_master(array)
        self.comm.barrier()

    def _read_master(self):
        with ulm.open(self._path) as reader:
            return reader.array

    def _write_master(self, array):
        assert array is not None
        with ulm.open(self._path, 'w') as writer:
            writer.write(array=array)


class FXCCorrelation:
    def __init__(self,
                 calc,
                 xc='RPA',
                 nlambda=8,
                 frequencies=None,
                 weights=None,
                 density_cut=1.e-6,
                 unit_cells=None,
                 tag=None,
                 avg_scheme=None,
                 *,
                 ecut,
                 **kwargs):

        self.ecut = ecut
        if isinstance(ecut, (float, int)):
            self.ecut_max = ecut
        else:
            self.ecut_max = max(ecut)

        self.rpa = RPACorrelation(
            calc,
            xc=xc,
            nlambda=nlambda,
            frequencies=frequencies,
            weights=weights,
            calculate_q=self.calculate_q_fxc,
            ecut=self.ecut,
            **kwargs)

        self.gs = self.rpa.gs
        self.context = self.rpa.context

        self.l_l, self.weight_l = p_roots(nlambda)
        self.l_l = (self.l_l + 1.0) * 0.5
        self.weight_l *= 0.5
        self.xc = xc
        self.density_cut = density_cut
        if unit_cells is None:
            unit_cells = self.gs.kd.N_c
        self.unit_cells = unit_cells

        self.xcflags = XCFlags(self.xc)
        self.avg_scheme = self.xcflags.choose_avg_scheme(avg_scheme)

        if tag is None:
            tag = self.gs.atoms.get_chemical_formula(mode='hill')
            if self.avg_scheme is not None:
                tag += '_' + self.avg_scheme

        self.cache = FXCCache(self.context.comm, tag, self.xc, self.ecut_max)

        self.omega_w = self.rpa.omega_w
        self.ibzq_qc = self.rpa.ibzq_qc
        self.nblocks = self.rpa.nblocks
        self.weight_w = self.rpa.weight_w

    @property
    def blockcomm(self):
        # Cannot be aliased as attribute
        # because rpa gets blockcomm during calculate
        return self.rpa.blockcomm

    def _calculate_kernel(self):
        # Find the first q vector to calculate kernel for
        # (density averaging scheme always calculates all q points anyway)

        q_empty = None

        for iq in reversed(range(len(self.ibzq_qc))):
            handle = self.cache.handle(iq)

            if not handle.exists():
                q_empty = iq

        if q_empty is None:
            self.context.print('%s kernel already calculated\n' %
                               self.xc)
            return

        kernelkwargs = dict(
            gs=self.gs,
            xc=self.xc,
            ibzq_qc=self.ibzq_qc,
            ecut=self.ecut_max,
            context=self.context)

        if self.avg_scheme == 'wavevector':
            self.context.print('Calculating %s kernel starting from '
                               'q point %s \n' % (self.xc, q_empty))
            kernelkwargs.update(q_empty=q_empty)
            kernel = KernelWave(**kernelkwargs)
        else:
            kernel = KernelDens(**kernelkwargs,
                                unit_cells=self.unit_cells,
                                density_cut=self.density_cut)

        for iq, fhxc_GG in kernel.calculate_fhxc():
            if self.context.comm.rank == 0:
                assert isinstance(fhxc_GG, np.ndarray), str(fhxc_GG)
            self.cache.handle(iq).write(fhxc_GG)

    @timer('FXC')
    def calculate(self, *, nbands=None):
        # kernel not required for RPA
        if self.xc != 'RPA':
            self._calculate_kernel()

        return self.rpa.calculate(spin=self.gs.nspins > 1, nbands=nbands)

    @timer('Chi0(q)')
    def calculate_q_fxc(self, chi0calc, chi0_s, m1, m2, gcut):
        for s, chi0 in enumerate(chi0_s):
            chi0calc.update_chi0(chi0, m1, m2, [s])

        self.context.print('E_c(q) = ', end='', flush=False)

        qpd = chi0.qpd
        nw = chi0.nw
        mynw = nw // self.nblocks
        assert nw % self.nblocks == 0
        nspins = len(chi0_s)
        nG = qpd.ngmax
        chi0_swGG = np.empty((nspins, mynw, nG, nG), complex)
        for chi0_wGG, chi0 in zip(chi0_swGG, chi0_s):
            chi0_wGG[:] = chi0.copy_array_with_distribution('wGG')
        if self.nblocks > 1:
            chi0_swGG = np.swapaxes(chi0_swGG, 2, 3)

        if not qpd.kd.gamma:
            e = self.calculate_energy_fxc(qpd, chi0_swGG, gcut)
            self.context.print('%.3f eV' % (e * Ha))
        else:
            W1 = self.blockcomm.rank * mynw
            W2 = W1 + mynw
            e = 0.0
            for v in range(3):
                for chi0_wGG, chi0 in zip(chi0_swGG, chi0_s):
                    chi0_wGG[:, 0] = chi0.chi0_WxvG[W1:W2, 0, v]
                    chi0_wGG[:, :, 0] = chi0.chi0_WxvG[W1:W2, 1, v]
                    chi0_wGG[:, 0, 0] = chi0.chi0_Wvv[W1:W2, v, v]
                ev = self.calculate_energy_fxc(qpd, chi0_swGG, gcut)
                e += ev
                self.context.print('%.3f' % (ev * Ha), end='', flush=False)
                if v < 2:
                    self.context.print('/', end='', flush=False)
                else:
                    self.context.print('eV')
            e /= 3

        return e

    def calculate_energy_contribution(self, chi0v_sGsG, fv_sGsG, nG):
        """Calculate contribution to energy from a single frequency point.

        The RPA correlation energy is the integral over all frequencies
        from 0 to infinity of this expression."""

        e = 0.0
        assert len(chi0v_sGsG) % nG == 0
        ns = len(chi0v_sGsG) // nG

        for l, weight in zip(self.l_l, self.weight_l):
            chiv = np.linalg.solve(
                np.eye(nG * ns) - l * np.dot(chi0v_sGsG, fv_sGsG),
                chi0v_sGsG).real  # this is SO slow

            chiv = chiv.reshape(ns, nG, ns, nG)
            for s1 in range(ns):
                for s2 in range(ns):
                    e -= np.trace(chiv[s1, :, s2, :]) * weight

        e += np.trace(chi0v_sGsG.real)
        return e

    @timer('Energy')
    def calculate_energy_fxc(self, qpd, chi0_swGG, gcut):
        """Evaluate correlation energy from chi0 and the kernel fhxc"""
        ibzq2_q = [
            np.dot(self.ibzq_qc[i] - qpd.q_c,
                   self.ibzq_qc[i] - qpd.q_c)
            for i in range(len(self.ibzq_qc))
        ]

        qi = np.argsort(ibzq2_q)[0]

        G_G = gcut.cut(qpd.G2_qG[0]**0.5)  # |G+q|

        nG = len(G_G)
        ns = len(chi0_swGG)

        # There are three options to calculate the
        # energy depending on kernel and/or averaging scheme.

        # Option (1) - Spin-polarized form of kernel exists
        #              e.g. rALDA, rAPBE.
        #              Then, solve block diagonal form of Dyson
        #              equation (dimensions (ns*nG) * (ns*nG))
        #              (note this does not necessarily mean that
        #              the calculation is spin-polarized!)

        if self.xcflags.spin_kernel:
            fv_GG = gcut.spin_cut(self.cache.handle(qi).read(), ns)

            # the spin-polarized kernel constructed from wavevector average
            # is already multiplied by |q+G| |q+G'|/4pi, and doesn't require
            # special treatment of the head and wings.  However not true for
            # density average:

            if self.avg_scheme == 'density':
                # Create and modify a view:
                fv_sGsG = fv_GG.reshape(ns, nG, ns, nG)

                for s1 in range(ns):
                    for s2 in range(ns):
                        fv_sGsG[s1, :, s2, :] *= (
                            G_G * G_G[:, np.newaxis] / (4 * np.pi))

                        if np.prod(self.unit_cells) > 1 and qpd.kd.gamma:
                            fv_sGsG[s1, 0, s2, :] = 0.0
                            fv_sGsG[s1, :, s2, 0] = 0.0
                            fv_sGsG[s1, 0, s2, 0] = 1.0

        else:
            fv_GG = np.eye(nG)

        if qpd.kd.gamma:
            G_G[0] = 1.0

        e_w = []

        # Loop over frequencies
        for chi0_sGG in np.swapaxes(chi0_swGG, 0, 1):
            chi0_sGG = gcut.cut(chi0_sGG, [1, 2])

            if self.xcflags.spin_kernel:
                chi0v_sGsG = get_chi0v_foreach_spin(chi0_sGG, G_G)
            else:
                chi0v_sGsG = get_chi0v_spinsum(chi0_sGG, G_G)

            energy = self.calculate_energy_contribution(
                chi0v_sGsG, fv_GG, nG)
            e_w.append(energy)

        E_w, energies = self.rpa.gather_energies(e_w)
        return energies


class KernelIntegrator(ABC):
    def __init__(self, gs, xc, context, ibzq_qc, ecut):
        self.gs = gs
        self.xc = xc
        self.context = context

        self.xcflags = XCFlags(xc)
        self.gd = gs.density.gd
        self.ibzq_qc = ibzq_qc
        self.ecut = ecut

    def calculate_fhxc(self):
        self.context.print(
            'Calculating {self.xc} kernel at {self.ecut} eV cutoff')
        fhxc_iterator = self._calculate_fhxc()

        while True:
            with self.context.timer('FHXC'):
                try:
                    yield next(fhxc_iterator)
                except StopIteration:
                    return

    @abstractmethod
    def _calculate_fhxc(self):
        """Perform computation and yield (iq, array) tuples."""


class KernelWave(KernelIntegrator):
    def __init__(self, *, q_empty, **kwargs):
        super().__init__(**kwargs)

        self.ns = self.gs.nspins
        self.q_empty = q_empty

        # Density grid
        n_sg, finegd = self.gs.get_all_electron_density(gridrefinement=2)
        self.n_g = n_sg.sum(axis=0).flatten()

        #  For atoms with large vacuum regions
        #  this apparently can take negative values!
        mindens = np.amin(self.n_g)

        if mindens < 0:
            self.context.print('Negative densities found! (magnitude %s)' %
                               np.abs(mindens), flush=False)
            self.context.print('These will be reset to 1E-12 elec/bohr^3)')
            self.n_g[np.where(self.n_g < 0.0)] = 1.0E-12

        r_g = finegd.get_grid_point_coordinates()
        self.x_g = 1.0 * r_g[0].flatten()
        self.y_g = 1.0 * r_g[1].flatten()
        self.z_g = 1.0 * r_g[2].flatten()
        self.gridsize = len(self.x_g)
        assert len(self.n_g) == self.gridsize

        # Enhancement factor for GGA
        if self.xcflags.is_apbe:
            nf_g = self.gs.hacky_all_electron_density(gridrefinement=4)
            gdf = self.gd.refine().refine()
            grad_v = [Gradient(gdf, v, n=1).apply for v in range(3)]
            gradnf_vg = gdf.empty(3)

            for v in range(3):
                grad_v[v](nf_g, gradnf_vg[v])

            self.s2_g = np.sqrt(np.sum(gradnf_vg[:, ::2, ::2, ::2]**2.0,
                                       0)).flatten()  # |\nabla\rho|
            self.s2_g *= 1.0 / (2.0 * (3.0 * np.pi**2.0)**(1.0 / 3.0) *
                                self.n_g**(4.0 / 3.0))
            # |\nabla\rho|/(2kF\rho) = s
            self.s2_g = self.s2_g**2  # s^2
            assert len(self.n_g) == len(self.s2_g)

            # Now we find all the regions where the
            # APBE kernel wants to be positive, and hack s to = 0,
            # so that we are really using the ALDA kernel
            # at these points
            apbe_g = get_pbe_fxc(self.n_g, self.s2_g)
            poskern_ind = np.where(apbe_g >= 0.0)
            if len(poskern_ind[0]) > 0:
                self.context.print(
                    'The APBE kernel takes positive values at '
                    + '%s grid points out of a total of %s (%3.2f%%).'
                    % (len(poskern_ind[0]), self.gridsize, 100.0 * len(
                        poskern_ind[0]) / self.gridsize), flush=False)
                self.context.print('The ALDA kernel will be used at these '
                                   'points')
                self.s2_g[poskern_ind] = 0.0

    def _calculate_fhxc(self):
        for iq, q_c in enumerate(self.ibzq_qc):
            if iq < self.q_empty:  # don't recalculate q vectors
                continue

            yield iq, self.calculate_one_qpoint(iq, q_c)

    def calculate_one_qpoint(self, iq, q_c):
        qpd = SingleQPWDescriptor.from_q(q_c, self.ecut / Ha, self.gd)

        nG = qpd.ngmax
        G_G = qpd.G2_qG[0]**0.5  # |G+q|
        Gv_G = qpd.get_reciprocal_vectors(q=0, add_q=False)
        # G as a vector (note we are at a specific q point here so set q=0)

        # Distribute G vectors among processors
        # Later we calculate for iG' > iG,
        # so stagger allocation in order to balance load
        local_Gvec_grid_size = nG // self.context.comm.size
        my_Gints = (self.context.comm.rank + np.arange(0,
                    local_Gvec_grid_size * self.context.comm.size,
                    self.context.comm.size))

        if (self.context.comm.rank +
                (local_Gvec_grid_size) * self.context.comm.size) < nG:
            my_Gints = np.append(my_Gints,
                                 [self.context.comm.rank +
                                  local_Gvec_grid_size *
                                  self.context.comm.size])

        my_Gv_G = Gv_G[my_Gints]

        # XXX Should this be if self.ns == 2 and self.xcflags.spin_kernel?
        calc_spincorr = (self.ns == 2) and (self.xc == 'rALDA'
                                            or self.xc == 'rAPBE')

        if calc_spincorr:
            # Form spin-dependent kernel according to
            # PRB 88, 115131 (2013) equation 20
            # (note typo, should be \tilde{f^rALDA})
            # spincorr is just the ALDA exchange kernel
            # with a step function (\equiv \tilde{f^rALDA})
            # fHxc^{up up}     = fHxc^{down down} = fv_nospin + fv_spincorr
            # fHxc^{up down}   = fHxc^{down up}   = fv_nospin - fv_spincorr
            fv_spincorr_GG = np.zeros((nG, nG), dtype=complex)

        fv_nospin_GG = np.zeros((nG, nG), dtype=complex)

        for iG, Gv in zip(my_Gints, my_Gv_G):  # loop over G vecs

            # For all kernels we
            # treat head and wings analytically
            if G_G[iG] > 1.0E-5:
                # Symmetrised |q+G||q+G'|, where iG' >= iG
                mod_Gpq = np.sqrt(G_G[iG] * G_G[iG:])

                # Phase factor \vec{G}-\vec{G'}
                deltaGv = Gv - Gv_G[iG:]

                if self.xc == 'rALDA':
                    # rALDA trick: the Hartree-XC kernel is exactly
                    # zero for densities below rho_min =
                    # min_Gpq^3/(24*pi^2),
                    # so we don't need to include these contributions
                    # in the Fourier transform

                    min_Gpq = np.amin(mod_Gpq)
                    rho_min = min_Gpq**3.0 / (24.0 * np.pi**2.0)
                    small_ind = np.where(self.n_g >= rho_min)
                elif self.xcflags.is_apbe:
                    # rAPBE trick: the Hartree-XC kernel
                    # is exactly zero at grid points where
                    # min_Gpq > cutoff wavevector

                    min_Gpq = np.amin(mod_Gpq)
                    small_ind = np.where(min_Gpq <= np.sqrt(
                        -4.0 * np.pi /
                        get_pbe_fxc(self.n_g, self.s2_g)))
                else:
                    small_ind = np.arange(self.gridsize)

                phase_Gpq = np.exp(
                    -1.0j *
                    (deltaGv[:, 0, np.newaxis] * self.x_g[small_ind] +
                     deltaGv[:, 1, np.newaxis] * self.y_g[small_ind] +
                     deltaGv[:, 2, np.newaxis] * self.z_g[small_ind]))

                def scaled_fHxc(spincorr):
                    return self.get_scaled_fHxc_q(
                        q=mod_Gpq,
                        sel_points=small_ind,
                        Gphase=phase_Gpq,
                        spincorr=spincorr)

                fv_nospin_GG[iG, iG:] = scaled_fHxc(spincorr=False)

                if calc_spincorr:
                    fv_spincorr_GG[iG, iG:] = scaled_fHxc(spincorr=True)
            else:
                # head and wings of q=0 are dominated by
                # 1/q^2 divergence of scaled Coulomb interaction

                assert iG == 0

                # The [0, 0] element would ordinarily be set to
                # 'l' if we have nonlinear kernel (which we are
                # removing).  Now l=1.0 always:
                fv_nospin_GG[0, 0] = 1.0
                fv_nospin_GG[0, 1:] = 0.0

                if calc_spincorr:
                    fv_spincorr_GG[0, :] = 0.0

            # End loop over G vectors

        self.context.comm.sum(fv_nospin_GG)

        # We've only got half the matrix here,
        # so add the hermitian conjugate:
        fv_nospin_GG += np.conj(fv_nospin_GG.T)
        # but now the diagonal's been doubled,
        # so we multiply these elements by 0.5
        fv_nospin_GG[np.diag_indices(nG)] *= 0.5

        # End of loop over coupling constant

        if calc_spincorr:
            self.context.comm.sum(fv_spincorr_GG)
            fv_spincorr_GG += np.conj(fv_spincorr_GG.T)
            fv_spincorr_GG[np.diag_indices(nG)] *= 0.5

        self.context.print('q point %s complete' % iq)

        if self.context.comm.rank == 0:
            if calc_spincorr:
                # Form the block matrix kernel
                fv_full_2G2G = np.empty((2 * nG, 2 * nG), dtype=complex)
                fv_full_2G2G[:nG, :nG] = fv_nospin_GG + fv_spincorr_GG
                fv_full_2G2G[:nG, nG:] = fv_nospin_GG - fv_spincorr_GG
                fv_full_2G2G[nG:, :nG] = fv_nospin_GG - fv_spincorr_GG
                fv_full_2G2G[nG:, nG:] = fv_nospin_GG + fv_spincorr_GG
                fhxc_sGsG = fv_full_2G2G

            else:
                fhxc_sGsG = fv_nospin_GG

            return fhxc_sGsG
        else:
            return None

    def get_scaled_fHxc_q(self, q, sel_points, Gphase, spincorr):
        # Given a coupling constant l, construct the Hartree-XC
        # kernel in q space a la Lein, Gross and Perdew,
        # Phys. Rev. B 61, 13431 (2000):
        #
        # f_{Hxc}^\lambda(q,\omega,r_s) = \frac{4\pi \lambda }{q^2}  +
        # \frac{1}{\lambda} f_{xc}(q/\lambda,\omega/\lambda^2,\lambda r_s)
        #
        # divided by the unscaled Coulomb interaction!!
        #
        # i.e. this subroutine returns f_{Hxc}^\lambda(q,\omega,r_s)
        #                              *  \frac{q^2}{4\pi}
        # = \lambda * [\frac{(q/lambda)^2}{4\pi}
        #              f_{Hxc}(q/\lambda,\omega/\lambda^2,\lambda r_s)]
        # = \lambda * [1/scaled_coulomb * fHxc computed with scaled quantities]

        # Apply scaling
        rho = self.n_g[sel_points]

        # GGA enhancement factor s is lambda independent,
        # but we might want to truncate it
        if self.xcflags.is_apbe:
            s2_g = self.s2_g[sel_points]
        else:
            s2_g = None

        l = 1.0  # Leftover from the age of non-linear kernels.
        # This would be an integration weight or something.
        scaled_q = q / l
        scaled_rho = rho / l**3.0
        scaled_rs = (3.0 / (4.0 * np.pi * scaled_rho))**(1.0 / 3.0
                                                         )  # Wigner radius

        if not spincorr:
            scaled_kernel = l * self.get_fHxc_q(scaled_rs, scaled_q, Gphase,
                                                s2_g)
        else:
            scaled_kernel = l * self.get_spinfHxc_q(scaled_rs, scaled_q,
                                                    Gphase, s2_g)

        return scaled_kernel

    def get_fHxc_q(self, rs, q, Gphase, s2_g):
        # Construct fHxc(q,G,:), divided by scaled Coulomb interaction

        heg = HEG(rs)
        qF = heg.qF

        fHxc_Gr = get_fHxc_Gr(self.xcflags, rs, q, qF, s2_g)

        # Integrate over r with phase
        fHxc_Gr *= Gphase
        fHxc_GG = np.sum(fHxc_Gr, 1) / self.gridsize
        return fHxc_GG

    def get_spinfHxc_q(self, rs, q, Gphase, s2_g):
        qF = HEG(rs).qF

        if self.xc == 'rALDA':
            fspinHxc_Gr = get_fspinHxc_Gr_rALDA(qF, q)

        elif self.xc == 'rAPBE':
            fspinHxc_Gr = get_fspinHxc_Gr_rAPBE(rs, q, s2_g)

        fspinHxc_Gr *= Gphase
        fspinHxc_GG = np.sum(fspinHxc_Gr, 1) / self.gridsize
        return fspinHxc_GG


class KernelDens(KernelIntegrator):
    def __init__(self, *, unit_cells, density_cut, **kwargs):
        super().__init__(**kwargs)

        self.unit_cells = unit_cells
        self.density_cut = density_cut

        self.A_x = -(3 / 4.) * (3 / np.pi)**(1 / 3.)

        self.n_g = self.gs.hacky_all_electron_density(gridrefinement=1)

        if self.xc[-3:] == 'PBE':
            nf_g = self.gs.hacky_all_electron_density(gridrefinement=2)
            gdf = self.gd.refine()
            grad_v = [Gradient(gdf, v, n=1).apply for v in range(3)]
            gradnf_vg = gdf.empty(3)
            for v in range(3):
                grad_v[v](nf_g, gradnf_vg[v])
            self.gradn_vg = gradnf_vg[:, ::2, ::2, ::2]

        qd = KPointDescriptor(self.ibzq_qc)
        self.pd = PWDescriptor(self.ecut / Ha, self.gd, complex, qd)

    def _calculate_fhxc(self):
        if self.xc[0] == 'r':  # wth?
            assert self.xcflags.spin_kernel
            yield from self.calculate_rkernel()
        else:
            assert self.xc[0] == 'A'  # wth?
            assert self.xc == 'ALDA'
            yield from self.calculate_local_kernel()

    def calculate_rkernel(self):
        gd = self.gd
        ng_c = gd.N_c
        cell_cv = gd.cell_cv
        icell_cv = 2 * np.pi * np.linalg.inv(cell_cv)
        vol = gd.volume

        ns = self.gs.nspins
        n_g = self.n_g  # density on rough grid

        fx_g = ns * self.get_fxc_g(n_g)  # local exchange kernel
        try:
            qc_g = (-4 * np.pi * ns / fx_g)**0.5  # cutoff functional
        except FloatingPointError as err:
            msg = (
                'Kernel is negative yet we want its square root.  '
                'You probably should not rely on this feature at all.  ',
                'See discussion https://gitlab.com/gpaw/gpaw/-/issues/723')
            raise RuntimeError(msg) from err
        flocal_g = qc_g**3 * fx_g / (6 * np.pi**2)  # ren. x-kernel for r=r'
        Vlocal_g = 2 * qc_g / np.pi  # ren. Hartree kernel for r=r'

        ng = np.prod(ng_c)  # number of grid points
        r_vg = gd.get_grid_point_coordinates()
        rx_g = r_vg[0].flatten()
        ry_g = r_vg[1].flatten()
        rz_g = r_vg[2].flatten()

        self.context.print('    %d grid points and %d plane waves at the '
                           'Gamma point' % (ng, self.pd.ngmax), flush=False)

        # Unit cells
        R_Rv = []
        weight_R = []
        nR_v = self.unit_cells
        nR = np.prod(nR_v)
        for i in range(-nR_v[0] + 1, nR_v[0]):
            for j in range(-nR_v[1] + 1, nR_v[1]):
                for h in range(-nR_v[2] + 1, nR_v[2]):
                    R_Rv.append(i * cell_cv[0] + j * cell_cv[1] +
                                h * cell_cv[2])
                    weight_R.append((nR_v[0] - abs(i)) * (nR_v[1] - abs(j)) *
                                    (nR_v[2] - abs(h)) / float(nR))
        if nR > 1:
            # with more than one unit cell only the exchange kernel is
            # calculated on the grid. The bare Coulomb kernel is added
            # in PW basis and Vlocal_g only the exchange part
            dv = self.gs.density.gd.dv
            gc = (3 * dv / 4 / np.pi)**(1 / 3.)
            Vlocal_g -= 2 * np.pi * gc**2 / dv
            self.context.print(
                '    Lattice point sampling: (%s x %s x %s)^2 '
                % (nR_v[0], nR_v[1], nR_v[2]) + ' Reduced to %s lattice points'
                % len(R_Rv), flush=False)

        l_g_size = -(-ng // self.context.comm.size)
        l_g_range = range(self.context.comm.rank * l_g_size,
                          min((self.context.comm.rank + 1) * l_g_size, ng))

        fhxc_qsGr = {}
        for iq in range(len(self.ibzq_qc)):
            fhxc_qsGr[iq] = np.zeros(
                (ns, len(self.pd.G2_qG[iq]), len(l_g_range)), dtype=complex)

        inv_error = np.seterr()
        np.seterr(invalid='ignore')
        np.seterr(divide='ignore')

        t0 = time()
        # Loop over Lattice points
        for i, R_v in enumerate(R_Rv):
            # Loop over r'. f_rr and V_rr are functions of r (dim. as r_vg[0])
            if i == 1:
                self.context.print(
                    '      Finished 1 cell in %s seconds' % int(time() - t0) +
                    ' - estimated %s seconds left' % int((len(R_Rv) - 1) *
                                                         (time() - t0)))
            if len(R_Rv) > 5:
                if (i + 1) % (len(R_Rv) / 5 + 1) == 0:
                    self.context.print(
                        '      Finished %s cells in %s seconds'
                        % (i, int(time() - t0)) + ' - estimated '
                        '%s seconds left' % int((len(R_Rv) - i) * (time() -
                                                                   t0) / i))
            for g in l_g_range:
                rx = rx_g[g] + R_v[0]
                ry = ry_g[g] + R_v[1]
                rz = rz_g[g] + R_v[2]

                # |r-r'-R_i|
                rr = ((r_vg[0] - rx)**2 + (r_vg[1] - ry)**2 +
                      (r_vg[2] - rz)**2)**0.5

                n_av = (n_g + n_g.flatten()[g]) / 2.
                fx_g = ns * self.get_fxc_g(n_av, index=g)
                qc_g = (-4 * np.pi * ns / fx_g)**0.5
                x = qc_g * rr
                osc_x = np.sin(x) - x * np.cos(x)
                f_rr = fx_g * osc_x / (2 * np.pi**2 * rr**3)
                if nR > 1:  # include only exchange part of the kernel here
                    V_rr = (sici(x)[0] * 2 / np.pi - 1) / rr
                else:  # include the full kernel (also hartree part)
                    V_rr = (sici(x)[0] * 2 / np.pi) / rr

                # Terms with r = r'
                if (np.abs(R_v) < 0.001).all():
                    tmp_flat = f_rr.flatten()
                    tmp_flat[g] = flocal_g.flatten()[g]
                    f_rr = tmp_flat.reshape(ng_c)
                    tmp_flat = V_rr.flatten()
                    tmp_flat[g] = Vlocal_g.flatten()[g]
                    V_rr = tmp_flat.reshape(ng_c)
                    del tmp_flat

                f_rr[np.where(n_av < self.density_cut)] = 0.0
                V_rr[np.where(n_av < self.density_cut)] = 0.0

                f_rr *= weight_R[i]
                V_rr *= weight_R[i]

                # r-r'-R_i
                r_r = np.array([r_vg[0] - rx, r_vg[1] - ry, r_vg[2] - rz])

                # Fourier transform of r
                for iq, q in enumerate(self.ibzq_qc):
                    q_v = np.dot(q, icell_cv)
                    e_q = np.exp(-1j * gemmdot(q_v, r_r, beta=0.0))
                    f_q = self.pd.fft((f_rr + V_rr) * e_q, iq) * vol / ng
                    fhxc_qsGr[iq][0, :, g - l_g_range[0]] += f_q
                    if ns == 2:
                        f_q = self.pd.fft(V_rr * e_q, iq) * vol / ng
                        fhxc_qsGr[iq][1, :, g - l_g_range[0]] += f_q

        self.context.comm.barrier()

        np.seterr(**inv_error)

        for iq, q in enumerate(self.ibzq_qc):
            npw = len(self.pd.G2_qG[iq])
            fhxc_sGsG = np.zeros((ns * npw, ns * npw), complex)
            # parallelize over PW below
            l_pw_size = -(-npw // self.context.comm.size)
            l_pw_range = range(self.context.comm.rank * l_pw_size,
                               min((self.context.comm.rank + 1) * l_pw_size,
                                   npw))

            if self.context.comm.size > 1:
                # redistribute grid and plane waves in fhxc_qsGr[iq]
                bg1 = BlacsGrid(self.context.comm, 1, self.context.comm.size)
                bg2 = BlacsGrid(self.context.comm, self.context.comm.size, 1)
                bd1 = bg1.new_descriptor(npw, ng, npw,
                                         -(-ng // self.context.comm.size))
                bd2 = bg2.new_descriptor(npw, ng,
                                         -(-npw // self.context.comm.size),
                                         ng)

                fhxc_Glr = np.zeros((len(l_pw_range), ng), dtype=complex)
                if ns == 2:
                    Koff_Glr = np.zeros((len(l_pw_range), ng), dtype=complex)

                r = Redistributor(bg1.comm, bd1, bd2)
                r.redistribute(fhxc_qsGr[iq][0], fhxc_Glr, npw, ng)
                if ns == 2:
                    r.redistribute(fhxc_qsGr[iq][1], Koff_Glr, npw, ng)
            else:
                fhxc_Glr = fhxc_qsGr[iq][0]
                if ns == 2:
                    Koff_Glr = fhxc_qsGr[iq][1]

            # Fourier transform of r'
            for iG in range(len(l_pw_range)):
                f_g = fhxc_Glr[iG].reshape(ng_c)
                f_G = self.pd.fft(f_g.conj(), iq) * vol / ng
                fhxc_sGsG[l_pw_range[0] + iG, :npw] = f_G.conj()
                if ns == 2:
                    v_g = Koff_Glr[iG].reshape(ng_c)
                    v_G = self.pd.fft(v_g.conj(), iq) * vol / ng
                    fhxc_sGsG[npw + l_pw_range[0] + iG, :npw] = v_G.conj()

            if ns == 2:  # f_00 = f_11 and f_01 = f_10
                fhxc_sGsG[:npw, npw:] = fhxc_sGsG[npw:, :npw]
                fhxc_sGsG[npw:, npw:] = fhxc_sGsG[:npw, :npw]

            self.context.comm.sum(fhxc_sGsG)
            fhxc_sGsG /= vol

            if self.context.comm.rank == 0:
                if nR > 1:  # add Hartree kernel evaluated in PW basis
                    Gq2_G = self.pd.G2_qG[iq]
                    if (q == 0).all():
                        Gq2_G = Gq2_G.copy()
                        Gq2_G[0] = 1.
                    vq_G = 4 * np.pi / Gq2_G
                    fhxc_sGsG += np.tile(np.eye(npw) * vq_G, (ns, ns))
                yield iq, fhxc_sGsG
            else:
                yield iq, None

    def calculate_local_kernel(self):
        # Standard ALDA exchange kernel
        # Use with care. Results are very difficult to converge
        # Sensitive to density_cut
        ns = self.gs.nspins
        gd = self.gd
        pd = self.pd
        cell_cv = gd.cell_cv
        icell_cv = 2 * np.pi * np.linalg.inv(cell_cv)
        vol = gd.volume

        fxc_sg = ns * self.get_fxc_g(ns * self.n_g)
        fxc_sg[np.where(self.n_g < self.density_cut)] = 0.0

        r_vg = gd.get_grid_point_coordinates()

        for iq in range(len(self.ibzq_qc)):
            Gvec_Gc = np.dot(pd.get_reciprocal_vectors(q=iq, add_q=False),
                             cell_cv / (2 * np.pi))
            npw = len(Gvec_Gc)
            l_pw_size = -(-npw // self.context.comm.size)
            l_pw_range = range(self.context.comm.rank * l_pw_size,
                               min((self.context.comm.rank + 1) * l_pw_size,
                                   npw))
            fhxc_sGsG = np.zeros((ns * npw, ns * npw), dtype=complex)
            for s in range(ns):
                for iG in l_pw_range:
                    for jG in range(npw):
                        fxc = fxc_sg[s].copy()
                        dG_c = Gvec_Gc[iG] - Gvec_Gc[jG]
                        dG_v = np.dot(dG_c, icell_cv)
                        dGr_g = gemmdot(dG_v, r_vg, beta=0.0)
                        ft_fxc = gd.integrate(np.exp(-1j * dGr_g) * fxc)
                        fhxc_sGsG[s * npw + iG, s * npw + jG] = ft_fxc

            self.context.comm.sum(fhxc_sGsG)
            fhxc_sGsG /= vol

            Gq2_G = self.pd.G2_qG[iq]
            if (self.ibzq_qc[iq] == 0).all():
                Gq2_G[0] = 1.
            vq_G = 4 * np.pi / Gq2_G
            fhxc_sGsG += np.tile(np.eye(npw) * vq_G, (ns, ns))

            yield iq, fhxc_sGsG

    def get_fxc_g(self, n_g, index=None):
        if self.xc[-3:] == 'LDA':
            return self.get_lda_g(n_g)
        elif self.xc[-3:] == 'PBE':
            return self.get_pbe_g(n_g, index=index)
        else:
            raise '%s kernel not recognized' % self.xc

    def get_lda_g(self, n_g):
        return (4. / 9.) * self.A_x * n_g**(-2. / 3.)

    def get_pbe_g(self, n_g, index=None):
        if index is None:
            gradn_vg = self.gradn_vg
        else:
            gradn_vg = self.gs.density.gd.empty(3)
            for v in range(3):
                gradn_vg[v] = (self.gradn_vg[v] +
                               self.gradn_vg[v].flatten()[index]) / 2

        kf_g = (3. * np.pi**2 * n_g)**(1 / 3.)
        s2_g = np.zeros_like(n_g)
        for v in range(3):
            axpy(1.0, gradn_vg[v]**2, s2_g)
        s2_g /= 4 * kf_g**2 * n_g**2

        from gpaw.xc.fxc_kernels import get_pbe_fxc
        return get_pbe_fxc(n_g, s2_g)


class XCFlags:
    _accepted_flags = {
        'RPA',
        'rALDA',
        'rAPBE',
        'ALDA'}

    _spin_kernels = {'rALDA', 'rAPBE', 'ALDA'}

    def __init__(self, xc):
        if xc not in self._accepted_flags:
            raise RuntimeError('%s kernel not recognized' % self.xc)

        self.xc = xc

    @property
    def spin_kernel(self):
        # rALDA/rAPBE are the only kernels which have spin-dependent forms
        return self.xc in self._spin_kernels

    @property
    def is_apbe(self):
        # If new GGA kernels are added, maybe there should be an
        # is_gga property.
        return self.xc in {'rAPBE'}

    def choose_avg_scheme(self, avg_scheme=None):
        xc = self.xc

        if self.spin_kernel:
            if avg_scheme is None:
                avg_scheme = 'density'
                # Two-point scheme default for rALDA and rAPBE

        if avg_scheme == 'density':
            assert self.spin_kernel, ('Two-point density average '
                                      'only implemented for rALDA and rAPBE')

        elif xc != 'RPA':
            avg_scheme = 'wavevector'
        else:
            avg_scheme = None

        return avg_scheme
